/* Copyright (C) 2007 Christian Dywan <christian at twotoasts dot de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. */

#include "browser.h"

// -- GTK+ signal handlers begin here

void on_action_window_new_activate(GtkAction* action, CBrowser* browser)
{
    browser_new(NULL);
}

void on_action_tab_new_activate(GtkAction* action, CBrowser* browser)
{
    browser_new(browser);
    update_browser_actions(browser);
}

void on_action_open_activate(GtkAction* action, CBrowser* browser)
{
    GtkWidget* dialog;
    dialog = gtk_file_chooser_dialog_new("Open file"
                                 , GTK_WINDOW(browser->window)
                                 , GTK_FILE_CHOOSER_ACTION_OPEN
                                 , GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL
                                 , GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT
                                 , NULL);
     gtk_window_set_icon_name(GTK_WINDOW(dialog), GTK_STOCK_OPEN);
     if(gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT)
     {
      gchar* sFilename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
      document_load_uri(get_current_document(browser), sFilename);
      g_free(sFilename);
     }
    gtk_widget_destroy(dialog);
}

void on_action_tab_close_activate(GtkAction* action, CBrowser* browser)
{
    document_close(get_current_document(browser), browser);
}

void on_action_window_close_activate(GtkAction* action, CBrowser* browser)
{
    gtk_widget_destroy(GTK_WIDGET(browser->window));
}

void on_action_quit_activate(GtkAction* action, CBrowser* browser)
{
    gtk_main_quit();
}

void on_action_find_activate(GtkAction* action, CBrowser* browser)
{
    if(GTK_WIDGET_VISIBLE(browser->findbox))
     gtk_widget_hide(browser->findbox);
    else
    {
     gtk_widget_show(browser->findbox);
     gtk_widget_grab_focus(GTK_WIDGET(browser->findbox_text));
    }
}

void on_action_find_next_activate(GtkAction* action, CBrowser* browser)
{
    if(!GTK_WIDGET_VISIBLE(browser->findbox))
     ; // FIXME: What if the findbox is hidden?
    const gchar* sText = gtk_entry_get_text(GTK_ENTRY(browser->findbox_text));
    const gboolean bCase = gtk_toggle_tool_button_get_active(
     GTK_TOGGLE_TOOL_BUTTON(browser->findbox_case));
    document_find(get_current_document(browser), sText, bCase, TRUE);
}

void on_action_find_previous_activate(GtkAction* action, CBrowser* browser)
{
    if(!GTK_WIDGET_VISIBLE(browser->findbox))
     ; // FIXME: What if the findbox is hidden?
    const gchar* sText = gtk_entry_get_text(GTK_ENTRY(browser->findbox_text));
    const gboolean bCase = gtk_toggle_tool_button_get_active(
     GTK_TOGGLE_TOOL_BUTTON(browser->findbox_case));
    document_find(get_current_document(browser), sText, bCase, FALSE);
}

static void on_toolbar_navigation_notify_style(GObject* object, GParamSpec* arg1
 , CBrowser* browser)
{
    if(config->iToolbarStyle == CONFIG_TOOLBAR_DEFAULT)
    {
     gtk_toolbar_set_style(GTK_TOOLBAR(browser->navibar)
     , config_to_toolbarstyle(config->iToolbarStyle));
    }
}

void on_action_toolbar_navigation_activate(GtkToggleAction* action, CBrowser* browser)
{
    config->bToolbarNavigation = gtk_toggle_action_get_active(action);
    sokoke_widget_set_visible(browser->navibar, config->bToolbarNavigation);
}

void on_action_toolbar_bookmarks_activate(GtkToggleAction* action, CBrowser* browser)
{
    config->bToolbarBookmarks = gtk_toggle_action_get_active(action);
    sokoke_widget_set_visible(browser->bookmarkbar, config->bToolbarBookmarks);
}

void on_action_toolbar_downloads_activate(GtkToggleAction* action, CBrowser* browser)
{
    /*config->bToolbarDownloads = gtk_toggle_action_get_active(action);
    sokoke_widget_set_visible(browser->downloadbar, config->bToolbarDownloads);*/
}

void on_action_toolbar_status_activate(GtkToggleAction* action, CBrowser* browser)
{
    config->bToolbarStatus = gtk_toggle_action_get_active(action);
    sokoke_widget_set_visible(browser->statusbar, config->bToolbarStatus);
}

void on_action_refresh_activate(GtkAction* action, CBrowser* browser)
{
    GdkModifierType state = (GdkModifierType)0;
    gint x, y; gdk_window_get_pointer(NULL, &x, &y, &state);
    document_refresh(get_current_document(browser), state & GDK_SHIFT_MASK);
}

void on_action_refresh_stop_activate(GtkAction* action, CBrowser* browser)
{
    gchar* sStockId; g_object_get(action, "stock-id", &sStockId, NULL);
    // Refresh or stop, depending on the stock id
    if(!strcmp(sStockId, GTK_STOCK_REFRESH))
    {
     GdkModifierType state = (GdkModifierType)0;
     gint x, y; gdk_window_get_pointer(NULL, &x, &y, &state);
     document_refresh(get_current_document(browser), state & GDK_SHIFT_MASK);
    }
    else
     document_stop(get_current_document(browser));
    g_free(sStockId);
}

void on_action_stop_activate(GtkAction* action, CBrowser* browser)
{
    document_stop(get_current_document(browser));
}

void on_action_zoom_in_activate(GtkAction* action, CBrowser* browser)
{
    const gfloat zoom = document_get_text_size(get_current_document(browser));
    document_set_text_size(get_current_document(browser), zoom + 0.1);
}

void on_action_zoom_out_activate(GtkAction* action, CBrowser* browser)
{
    const gfloat zoom = document_get_text_size(get_current_document(browser));
    document_set_text_size(get_current_document(browser), zoom - 0.1);
}

void on_action_zoom_normal_activate(GtkAction* action, CBrowser* browser)
{
    document_set_text_size(get_current_document(browser), 1);
}

void on_action_back_activate(GtkAction* action, CBrowser* browser)
{
    document_go_back(get_current_document(browser));
}

void on_action_forward_activate(GtkAction* action, CBrowser* browser)
{
    document_go_forward(get_current_document(browser));
}

void on_action_home_activate(GtkAction* action, CBrowser* browser)
{
    document_load_uri(get_current_document(browser), config->sHomepage);
}

void on_action_location_activate(GtkAction* action, CBrowser* browser)
{
    if(GTK_WIDGET_VISIBLE(browser->navibar))
     gtk_widget_grab_focus(browser->location);
    else
    {
     // TODO: We should offer all of the toolbar location's features here
     GtkWidget* dialog;
     dialog = gtk_dialog_new_with_buttons("Open location"
                                 , GTK_WINDOW(browser->window)
                                 , GTK_DIALOG_DESTROY_WITH_PARENT
                                 | GTK_DIALOG_NO_SEPARATOR
                                 , GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL
                                 , GTK_STOCK_JUMP_TO, GTK_RESPONSE_ACCEPT
                                 , NULL);
     gtk_window_set_icon_name(GTK_WINDOW(dialog), GTK_STOCK_JUMP_TO);
     gtk_container_set_border_width(GTK_CONTAINER(dialog), 5);
     gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), 5);
     GtkWidget* hbox = gtk_hbox_new(FALSE, 8);
     gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);
     GtkWidget* label = gtk_label_new_with_mnemonic("_Location:");
     gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
     GtkWidget* entry = gtk_entry_new();
     gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);
     gtk_box_pack_start(GTK_BOX(hbox), entry, FALSE, FALSE, 0);
     gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), hbox);
     gtk_widget_show_all(hbox);
     gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);
     if(gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT)
     {
      gtk_entry_set_text(GTK_ENTRY(browser->location)
       , gtk_entry_get_text(GTK_ENTRY(entry)));
      GdkEventKey event;
      event.keyval = GDK_Return;
      on_location_key_down(browser->location
       , &event, browser);
     }
     gtk_widget_destroy(dialog);
    }
}

void on_action_websearch_activate(GtkAction* action, CBrowser* browser)
{
    if(GTK_WIDGET_VISIBLE(browser->webSearch)
     && GTK_WIDGET_VISIBLE(browser->navibar))
     gtk_widget_grab_focus(browser->webSearch);
    else
    {
     // TODO: We should offer all of the toolbar search's features here
     GtkWidget* dialog;
     dialog = gtk_dialog_new_with_buttons("Web search"
                                 , GTK_WINDOW(browser->window)
                                 , GTK_DIALOG_DESTROY_WITH_PARENT
                                 | GTK_DIALOG_NO_SEPARATOR
                                 , GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL
                                 , GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT
                                 , NULL);
     gtk_window_set_icon_name(GTK_WINDOW(dialog), GTK_STOCK_FIND);
     gtk_container_set_border_width(GTK_CONTAINER(dialog), 5);
     gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), 5);
     GtkWidget* hbox = gtk_hbox_new(FALSE, 8);
     gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);
     GtkWidget* label = gtk_label_new_with_mnemonic("_Location:");
     gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
     GtkWidget* entry = gtk_entry_new();
     gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);
     gtk_box_pack_start(GTK_BOX(hbox), entry, FALSE, FALSE, 0);
     gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), hbox);
     gtk_widget_show_all(hbox);
     gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);
     if(gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT)
     {
      gtk_entry_set_text(GTK_ENTRY(browser->webSearch)
       , gtk_entry_get_text(GTK_ENTRY(entry)));
      on_webSearch_activate(browser->webSearch, browser);
     }
     gtk_widget_destroy(dialog);
    }
}

void on_menu_tabsclosed_item_activate(GtkMenuItem* menuitem, CBrowser* browser)
{
    // Create a new document with an uri which has been closed before
    const gchar* sUri = (const gchar*)g_object_get_data(G_OBJECT(menuitem), "uri");
    CBrowser* curBrowser = browser_new(browser);
    document_load_uri(curBrowser->document, sUri);
    gtk_widget_destroy(GTK_WIDGET(menuitem));
    update_browser_actions(curBrowser);
}

void on_action_tabclosed_undo_activate(GtkAction* action, CBrowser* browser)
{
    // Emulate a click on the most recent closedtabs item
    GtkWidget* menuitem;
    GList* children = gtk_container_get_children(
     GTK_CONTAINER(browser->menu_tabsclosed));
    menuitem = (GtkWidget*)g_list_nth_data(children, 0);
    on_menu_tabsclosed_item_activate(GTK_MENU_ITEM(menuitem), browser);
    g_list_free(children);
}

void on_action_tabsclosed_clear_activate(GtkAction* action, CBrowser* browser)
{
    // Clear the closed tabs list, ie. all items that have an uri
    GtkWidget* menuitem; GtkWidget* icon;
    GList* children = gtk_container_get_children(
     GTK_CONTAINER(browser->menu_tabsclosed));
    do
    {
     menuitem = (GtkWidget*)children->data;
     if(g_object_get_data(G_OBJECT(menuitem), "uri"))
      gtk_widget_destroy(menuitem);
    }
    while(children = g_list_next(children)); g_list_free(children);
    update_browser_actions(browser);
}

void on_action_link_tab_new_activate(GtkAction* action, CBrowser* browser)
{
    CBrowser* curBrowser = browser_new(browser);
    document_load_uri(curBrowser->document, browser->sElementUri);
}

void on_action_link_tab_current_activate(GtkAction* action, CBrowser* browser)
{
    document_load_uri(get_current_document(browser), browser->sElementUri);
}

void on_action_link_window_new_activate(GtkAction* action, CBrowser* browser)
{
    CBrowser* curBrowser = browser_new(NULL);
    document_load_uri(curBrowser->document, browser->sElementUri);
}

void on_action_link_savewith_activate(GtkAction* action, CBrowser* browser)
{
    spawn_protocol_command("download", browser->sElementUri);
}

void on_action_link_copy_activate(GtkAction* action, CBrowser* browser)
{
    GtkClipboard* clipboard = gtk_clipboard_get(GDK_SELECTION_CLIPBOARD);
    gtk_clipboard_set_text(clipboard, browser->sElementUri, -1);
}

void on_action_colorpicker_activate(GtkAction* action, CBrowser* browser)
{
    static GtkWidget* dialog;
    // Show the color picker. Create it if necessary.
    if(GTK_IS_DIALOG(dialog))
     gtk_window_present(GTK_WINDOW(dialog));
    else
    {
     dialog = gtk_dialog_new_with_buttons("Color picker"
                                 , GTK_WINDOW(browser->window)
                                 , GTK_DIALOG_DESTROY_WITH_PARENT
                                 , GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE
                                 , NULL);
     gtk_window_set_icon_name(GTK_WINDOW(dialog), GTK_STOCK_COLOR_PICKER);
     g_signal_connect(GTK_OBJECT(dialog), "response"
      , GTK_SIGNAL_FUNC(gtk_widget_destroy), dialog);
     gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), 5);
     GtkWidget* color_selection = gtk_color_selection_new();
     gtk_container_set_border_width(GTK_CONTAINER(color_selection), 5);
     // TODO: Could we do anything to make the opacity useful?
     gtk_color_selection_set_has_opacity_control(
      GTK_COLOR_SELECTION(color_selection), FALSE);
     // TODO: Is the palette any good?
     gtk_color_selection_set_has_palette(
      GTK_COLOR_SELECTION(color_selection), TRUE);
     gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox)
      , color_selection, TRUE, TRUE, 15);
     gtk_box_set_spacing(GTK_BOX(GTK_DIALOG(dialog)->vbox), 2);
     // TODO: Any useful buttons we could provide?
     gtk_widget_show_all(dialog);
    }
}

void on_action_tab_previous_activate(GtkAction* action, CBrowser* browser)
{
    gint page = gtk_notebook_get_current_page(GTK_NOTEBOOK(browser->documents));
    gtk_notebook_set_current_page(GTK_NOTEBOOK(browser->documents), page - 1);
}

void on_action_tab_next_activate(GtkAction* action, CBrowser* browser)
{
    // Advance one tab or jump to the first one if we are at the last one
    gint page = gtk_notebook_get_current_page(GTK_NOTEBOOK(browser->documents));
    if(page == gtk_notebook_get_n_pages(GTK_NOTEBOOK(browser->documents)) - 1)
     page = -1;
    gtk_notebook_set_current_page(GTK_NOTEBOOK(browser->documents), page + 1);
}

void on_window_menu_item_activate(GtkImageMenuItem* widget, CBrowser* browser)
{
    gint page = get_document_index(browser->document, browser);
    gtk_notebook_set_current_page(GTK_NOTEBOOK(browser->documents), page);
}

void on_action_about_activate(GtkAction* action, CBrowser* browser)
{
    gtk_show_about_dialog(GTK_WINDOW(browser->window)
     , "logo-icon-name", gtk_window_get_icon_name(GTK_WINDOW(browser->window))
     , "name", PACKAGE_NAME
     , "version", PACKAGE_VERSION
     , "comments", "A lightweight web browser."
     , "copyright", "Copyright © 2007 Christian Dywan"
     , "website", "http://software.twotoasts.de"
     , "authors", credits_authors
     , "documenters", credits_documenters
     , "artists", credits_artists
     , "license", license_gpl_2
     , "wrap-license", TRUE
     //, "translator-credits", _("translator-credits")
     , NULL);
}

gboolean on_location_key_down(GtkWidget* widget
 , GdkEventKey* event, CBrowser* browser)
{
    const gchar* sUri; CBrowser* curBrowser;
    switch(event->keyval)
    {
     case GDK_Return:
      sUri = gtk_entry_get_text(GTK_ENTRY(widget));
      if(sUri)
      {
       gchar* sNewUri = magic_uri(sUri, TRUE);
       // TODO: Use sUrl intermediately when completion is better
       /* TODO Completion should be generated from history, that is
       the uri as well as the title. */
       entry_completion_append(GTK_ENTRY(widget), sUri);
       document_load_uri(get_current_document(browser), sNewUri);
       g_free(sNewUri);
      }
      return TRUE;
     case GDK_Escape:
      sUri = document_get_uri(get_current_document(browser));
      if(sUri && *sUri)
       gtk_entry_set_text(GTK_ENTRY(widget), sUri);
      return TRUE;
    }
    return FALSE;
}

void on_location_changed(GtkWidget* widget, CBrowser* browser)
{
    // Preserve changes to the uri
    const gchar* sNewUri = gtk_entry_get_text(GTK_ENTRY(widget));
    g_object_set_data(G_OBJECT(get_current_document(browser)), "uri", g_strdup(sNewUri));
}

void on_navibar_tabsclosed_clicked(GtkToolItem* widget, CBrowser* browser)
{
    gtk_menu_popup(GTK_MENU(browser->menu_tabsclosed), NULL, NULL, NULL, NULL, 0
     , gtk_get_current_event_time());
}

void on_action_panels_activate(GtkToggleAction* action, CBrowser* browser)
{
    config->bPanels = gtk_toggle_action_get_active(action);
    sokoke_widget_set_visible(browser->panels, config->bPanels);
}

void on_action_panel_item_activate(GtkRadioAction* action
 , GtkRadioAction* currentAction, CBrowser* browser)
{
    g_return_if_fail(GTK_IS_ACTION(action));
    // TODO: Activating again should hide the contents; how?
    //gint iValue; gint iCurrentValue;
    //g_object_get(G_OBJECT(action), "value", &iValue, NULL);
    //g_object_get(G_OBJECT(currentAction), "value", &iCurrentValue, NULL);
    //GtkWidget* parent = gtk_widget_get_parent(browser->panels_notebook);
    //sokoke_widget_set_visible(parent, iCurrentValue == iValue);
    /*gtk_paned_set_position(GTK_PANED(gtk_widget_get_parent(browser->panels))
     , iCurrentValue == iValue ? config->iPanelPos : 0);*/
    config->iPanelActive = gtk_radio_action_get_current_value(action);
    gint iPage = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(currentAction), "iPage"));
    gtk_notebook_set_current_page(GTK_NOTEBOOK(browser->panels_notebook), iPage);
    gtk_widget_show(browser->panels);
}

void on_action_openinpanel_activate(GtkAction* action, CBrowser* browser)
{
    GtkWidget* document = get_current_document(browser);
    g_free(config->sPanelPageholder);
    config->sPanelPageholder = g_strdup(document_get_uri(document));
    GtkAction* action_pageholder =
     gtk_action_group_get_action(browser->actiongroup, "PanelPageholder");
    gint value; g_object_get(G_OBJECT(action_pageholder), "value", &value, NULL);
    sokoke_radio_action_set_current_value(GTK_RADIO_ACTION(action_pageholder), value);
    gtk_widget_show(browser->panels);
    document_load_uri(browser->panel_pageholder, config->sPanelPageholder);
}


static void on_panels_notify_position(GObject* object, GParamSpec* arg1
 , CBrowser* browser)
{
    config->iWinPanelPos = gtk_paned_get_position(GTK_PANED(object));
}

void on_panels_button_close_clicked(GtkWidget* widget, CBrowser* browser)
{
    config->bPanels = FALSE;
    gtk_widget_hide(browser->panels);
}

gboolean on_notebook_tab_mouse_up(GtkWidget* widget, GdkEventButton* event
 , CBrowser* browser)
{
    if(event->button == 1 && event->type == GDK_2BUTTON_PRESS)
    {
     // Toggle the label visibility on double click
     GtkWidget* child = gtk_bin_get_child(GTK_BIN(widget));
     GList* children = gtk_container_get_children(GTK_CONTAINER(child));
     child = (GtkWidget*)g_list_nth_data(children, 1);
     gboolean visible = gtk_widget_get_child_visible(GTK_WIDGET(child));
     gtk_widget_set_child_visible(GTK_WIDGET(child), !visible);
     gint a, b; sokoke_widget_get_text_size(browser->document_name, "M", &a, &b);
     gtk_widget_set_size_request(child, !visible
      ? a * config->iTabSize : 0, !visible ? -1 : 0);
     g_list_free(children);
     return TRUE;
    }
    else if(event->button == 2)
    {
     // Close the document on middle click
     document_close(browser->document, browser);
     return TRUE;
    }

    return FALSE;
}

gboolean on_notebook_tab_close_clicked(GtkWidget* widget, CBrowser* browser)
{
    document_close(browser->document, browser);

    return TRUE;
}

void on_notebook_switch_page(GtkWidget* widget, GtkNotebookPage* page
 , guint page_num, CBrowser* browser)
{
    GtkWidget* docbox = gtk_notebook_get_nth_page(
     GTK_NOTEBOOK(browser->documents), page_num);
    GtkWidget* document = gtk_bin_get_child(GTK_BIN(docbox));
    const gchar* sUri = g_object_get_data(G_OBJECT(document), "uri");
    if(!sUri) sUri = document_get_uri(document);
    gtk_entry_set_text(GTK_ENTRY(browser->location), sUri);
    gchar* sTitle = g_object_get_data(G_OBJECT(document), "title");
    gchar* sWindowTitle = g_strconcat(sTitle, " - ", PACKAGE_NAME, NULL);
    gtk_window_set_title(GTK_WINDOW(browser->window), sWindowTitle);
    g_free(sWindowTitle);
    update_favicon(browser);
    update_security(browser);
    update_gui_state(browser);
    update_statusbar_text(browser);
    update_feeds(browser);
    update_search_engines(browser);
}

void on_findbox_button_close_clicked(GtkWidget* widget, CBrowser* browser)
{
    gtk_widget_hide(browser->findbox);
}

static gboolean on_window_configure(GtkWidget* widget, GdkEventConfigure* event
 , CBrowser* browser)
{
    gtk_window_get_position(GTK_WINDOW(browser->window)
     , &config->iWinLeft, &config->iWinTop);
    return FALSE;
}

static void on_window_size_allocate(GtkWidget* widget, GtkAllocation* allocation
 , CBrowser* browser)
{
    config->iWinWidth = allocation->width;
    config->iWinHeight = allocation->height;
}

gboolean on_window_destroy(GtkWidget* widget, GdkEvent* event, CBrowser* browser)
{
    gboolean bProceed = TRUE;
    // TODO: What if there are multiple windows?
    // TODO: Smart dialog, à la 'Session?: Save, Discard, Cancel'
    // TODO: Pref startup: session, ask, homepage, blank <-- ask
    // TODO: Pref quit: session, ask, none <-- ask

    if(0 /*GPOINTER_TO_INT(browser_count) > 1*/)
    {
     GtkDialog* dialog;
     dialog = GTK_DIALOG(gtk_message_dialog_new(GTK_WINDOW(browser->window)
      , GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_WARNING, GTK_BUTTONS_YES_NO
      , "There is more than one tab open. Do you want to close anyway?"));
     gtk_window_set_title(GTK_WINDOW(dialog), PACKAGE_NAME);
     gtk_dialog_set_default_response(dialog, GTK_RESPONSE_YES);
     bProceed = gtk_dialog_run(dialog) == GTK_RESPONSE_YES;
     gtk_widget_destroy(GTK_WIDGET(dialog));
    }
    return !bProceed;
}

// -- Browser creation begins here

static void create_bookmark_menu(XBELitem*, GtkWidget*, CBrowser*);

static void on_bookmark_menu_folder_activate(GtkWidget* menuitem, CBrowser* browser)
{
    GtkWidget* menu = gtk_menu_item_get_submenu(GTK_MENU_ITEM(menuitem));
    gtk_container_foreach(GTK_CONTAINER(menu), (GtkCallback)gtk_widget_destroy, NULL);//...
    XBELitem* folder = (XBELitem*)g_object_get_data(G_OBJECT(menuitem), "xbelitem");
    create_bookmark_menu(folder, menu, browser);
    // Remove all menuitems when the menu is hidden.
    // FIXME: We really *want* the line below, but it won't work like that
    //g_signal_connect_after(menu, "hide", G_CALLBACK(gtk_container_foreach), gtk_widget_destroy);
    gtk_widget_show(menuitem);
}

static void on_bookmark_toolbar_folder_activate(GtkToolItem* toolitem, CBrowser* browser)
{
    GtkWidget* menu = gtk_menu_new();
    XBELitem* folder = (XBELitem*)g_object_get_data(G_OBJECT(toolitem), "xbelitem");
    create_bookmark_menu(folder, menu, browser);
    // Remove all menuitems when the menu is hidden.
    // FIXME: We really *should* run the line below, but it won't work like that
    //g_signal_connect(menu, "hide", G_CALLBACK(gtk_container_foreach), gtk_widget_destroy);
    sokoke_widget_popup(GTK_WIDGET(toolitem), GTK_MENU(menu), NULL);
}

void on_menu_bookmarks_item_activate(GtkWidget* widget, CBrowser* browser)
{
    XBELitem* item = (XBELitem*)g_object_get_data(G_OBJECT(widget), "xbelitem");
    document_load_uri(get_current_document(browser), xbel_bookmark_get_href(item));
}

static void create_bookmark_menu(XBELitem* folder, GtkWidget* menu, CBrowser* browser)
{
    guint i; for(i = 0; i < xbel_folder_get_n_items(folder); i++)
    {
     XBELitem* item = xbel_folder_get_nth_item(folder, i);
     // FIXME: A separator has none of these two
     const gchar* sTitle = xbel_item_get_title(item);
     //const gchar* sDesc = xbel_item_get_desc(item);
     GtkWidget* menuitem;
     switch(xbel_item_get_type(item))
     {
      case XBEL_ITEM_FOLDER:
       menuitem = gtk_image_menu_item_new_with_label(sTitle);
        gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(menuitem)
         , gtk_image_new_from_stock(GTK_STOCK_DIRECTORY, GTK_ICON_SIZE_MENU));
       GtkWidget* _menu = gtk_menu_new();
       gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem), _menu);
       g_signal_connect(menuitem, "activate"
        , G_CALLBACK(on_bookmark_menu_folder_activate), browser);
       g_object_set_data(G_OBJECT(menuitem), "xbelitem", item);
       break;
      case XBEL_ITEM_BOOKMARK:
       menuitem = menu_item_new(sTitle, STOCK_BOOKMARK
        , G_CALLBACK(on_menu_bookmarks_item_activate), TRUE, browser);
       g_object_set_data(G_OBJECT(menuitem), "xbelitem", item);
       break;
      case XBEL_ITEM_SEPARATOR:
       menuitem = gtk_separator_menu_item_new();
       break;
      default:
       g_warning("Unknown xbel item type");
     }
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menuitem);
     gtk_widget_show(menuitem);
    }
}

CBrowser* browser_new(CBrowser* oldBrowser)
{
    browser_count = GUINT_TO_POINTER(GPOINTER_TO_UINT(browser_count) + 1);
    CBrowser* browser = g_new0(CBrowser, 1);
    browser_list = g_list_prepend(browser_list, browser);

    if(!oldBrowser) {

    GtkWidget* label; GtkWidget* hbox;

    // Setup the window metrics
    browser->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    GdkScreen* screen = gtk_window_get_screen(GTK_WINDOW(browser->window));
    const gint iDefaultWidth = (gint)gdk_screen_get_width(screen) / 1.7;
    const gint iDefaultHeight = (gint)gdk_screen_get_height(screen) / 1.7;
    if(config->bRememberWinMetrics)
    {
     if(!config->iWinWidth && !config->iWinHeight)
     {
      config->iWinWidth = iDefaultWidth;
      config->iWinHeight = iDefaultWidth;
     }
     if(config->iWinLeft && config->iWinTop)
      gtk_window_move(GTK_WINDOW(browser->window), config->iWinLeft, config->iWinTop);
     gtk_window_set_default_size(GTK_WINDOW(browser->window)
      , config->iWinWidth, config->iWinHeight);
    }
    else
     gtk_window_set_default_size(GTK_WINDOW(browser->window)
      , iDefaultWidth, iDefaultHeight);
    g_signal_connect(G_OBJECT(browser->window), "configure-event"
     , G_CALLBACK(on_window_configure), browser);
    g_signal_connect(G_OBJECT(browser->window), "size-allocate"
     , G_CALLBACK(on_window_size_allocate), browser);
    // FIXME: Use custom program icon
    gtk_window_set_icon_name(GTK_WINDOW(browser->window), "web-browser");
    gtk_window_set_title(GTK_WINDOW(browser->window), PACKAGE_NAME);
    gtk_window_add_accel_group(GTK_WINDOW(browser->window), accel_group);
    g_signal_connect(GTK_OBJECT(browser->window), "delete-event"
     , G_CALLBACK(on_window_destroy), browser);
    GtkWidget* vbox = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(browser->window), vbox);
    gtk_widget_show(vbox);

    // Let us see some ui manager magic
    browser->actiongroup = gtk_action_group_new("Browser");
    gtk_action_group_add_actions(browser->actiongroup, entries, entries_n, browser);
    gtk_action_group_add_toggle_actions(browser->actiongroup
     , toggle_entries, toggle_entries_n, browser);
    gtk_action_group_add_radio_actions(browser->actiongroup
     , refreshevery_entries, refreshevery_entries_n
     , 300, NULL/*G_CALLBACK(activate_refreshevery_period_action)*/, browser);
    gtk_action_group_add_radio_actions(browser->actiongroup
     , panel_entries, panel_entries_n, -1
     , G_CALLBACK(on_action_panel_item_activate), browser);
    GtkUIManager* ui_manager = gtk_ui_manager_new();
    gtk_ui_manager_insert_action_group(ui_manager, browser->actiongroup, 0);
    gtk_window_add_accel_group(GTK_WINDOW(browser->window)
     , gtk_ui_manager_get_accel_group(ui_manager));

    GError* error = NULL;
    if(!gtk_ui_manager_add_ui_from_string(ui_manager, ui_markup, -1, &error))
    {
     // TODO: Should this be a message dialog? When does this happen?
     g_message("User interface couldn't be created: %s", error->message);
     g_error_free(error);
    }

    GtkAction* action;
    // Make all actions except toplevel menus which lack a callback insensitive
    // This will vanish once all actions are implemented
    gint i; for(i = 0; i < entries_n; i++)
    {
     action = gtk_action_group_get_action(browser->actiongroup, entries[i].name);
     gtk_action_set_sensitive(action, entries[i].callback || !entries[i].tooltip);
    }
    for(i = 0; i < toggle_entries_n; i++)
    {
     action = gtk_action_group_get_action(browser->actiongroup
      , toggle_entries[i].name);
     gtk_action_set_sensitive(action, toggle_entries[i].callback != NULL);
    }
    for(i = 0; i < refreshevery_entries_n; i++)
    {
     action = gtk_action_group_get_action(browser->actiongroup
      , refreshevery_entries[i].name); gtk_action_set_sensitive(action, FALSE);
    }

    action_set_sensitive("PrivateBrowsing", FALSE, browser);
    action_set_sensitive("WorkOffline", FALSE, browser);

    action_set_active("ToolbarNavigation", config->bToolbarNavigation, browser);
    action_set_active("ToolbarBookmarks", config->bToolbarBookmarks, browser);
    //action_set_active("ToolbarDownloads", config->bToolbarDownloads, browser);
    action_set_active("ToolbarStatus", config->bToolbarStatus, browser);
    action_set_active("Panels", config->bPanels, browser);
    action = gtk_action_group_get_action(browser->actiongroup, "PanelDownloads");
    sokoke_radio_action_set_current_value(GTK_RADIO_ACTION(action), config->iPanelActive);

    // Create the menubar
    browser->menubar = gtk_ui_manager_get_widget(ui_manager, "/menubar");
    GtkWidget* menuitem = gtk_menu_item_new();
    gtk_widget_show(menuitem);
    browser->throbber = thunar_throbber_new();
    gtk_widget_show(browser->throbber);
    gtk_container_add(GTK_CONTAINER(menuitem), browser->throbber);
    gtk_widget_set_sensitive(menuitem, FALSE);
    gtk_menu_item_set_right_justified(GTK_MENU_ITEM(menuitem), TRUE);
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menubar), menuitem);
    gtk_box_pack_start(GTK_BOX(vbox), browser->menubar, FALSE, FALSE, 0);
    browser->menu_tabsclosed = gtk_ui_manager_get_widget(ui_manager, "/popup_tabsclosed");
    menuitem = gtk_ui_manager_get_widget(ui_manager, "/menubar/Go/TabsClosed");
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem), browser->menu_tabsclosed);
    menuitem = gtk_separator_menu_item_new();
    gtk_widget_show(menuitem);
    gtk_menu_shell_prepend(GTK_MENU_SHELL(browser->menu_tabsclosed), menuitem);
    browser->menu_bookmarks = gtk_menu_item_get_submenu(
     GTK_MENU_ITEM(gtk_ui_manager_get_widget(ui_manager, "/menubar/Bookmarks")));
    menuitem = gtk_separator_menu_item_new();
    gtk_widget_show(menuitem);
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menu_bookmarks), menuitem);
    browser->menu_window = gtk_menu_item_get_submenu(
     GTK_MENU_ITEM(gtk_ui_manager_get_widget(ui_manager, "/menubar/Window")));
    menuitem = gtk_separator_menu_item_new();
    gtk_widget_show(menuitem);
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menu_window), menuitem);
    gtk_widget_show(browser->menubar);
    browser->popup_document = gtk_ui_manager_get_widget(ui_manager, "/popup_document");
    g_object_ref(browser->popup_document);
    browser->popup_element = gtk_ui_manager_get_widget(ui_manager, "/popup_element");
    g_object_ref(browser->popup_element);

    // Create the navigation toolbar
    browser->navibar = gtk_ui_manager_get_widget(ui_manager, "/toolbar_navigation");
    gtk_toolbar_set_style(GTK_TOOLBAR(browser->navibar)
     , config_to_toolbarstyle(config->iToolbarStyle));
    g_signal_connect(gtk_settings_get_default(), "notify::gtk-toolbar-style"
     , G_CALLBACK(on_toolbar_navigation_notify_style), browser);
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(browser->navibar)
     , config_to_toolbariconsize(config->bToolbarSmall));
    gtk_toolbar_set_show_arrow(GTK_TOOLBAR(browser->navibar), TRUE);
    gtk_box_pack_start(GTK_BOX(vbox), browser->navibar, FALSE, FALSE, 0);
    browser->newTab = gtk_ui_manager_get_widget(ui_manager, "/toolbar_navigation/TabNew");
    action = gtk_action_group_get_action(browser->actiongroup, "Back");
    g_object_set(action, "is-important", TRUE, NULL);

    // Location entry
    browser->location = sexy_icon_entry_new();
    entry_setup_completion(GTK_ENTRY(browser->location));
    sokoke_entry_set_can_undo(GTK_ENTRY(browser->location), TRUE);
    browser->location_icon = gtk_image_new();
    sexy_icon_entry_set_icon(SEXY_ICON_ENTRY(browser->location)
     , SEXY_ICON_ENTRY_PRIMARY, GTK_IMAGE(browser->location_icon));
    sexy_icon_entry_add_clear_button(SEXY_ICON_ENTRY(browser->location));
    g_signal_connect(browser->location, "key-press-event"
     , G_CALLBACK(on_location_key_down), browser);
    g_signal_connect(browser->location, "changed"
     , G_CALLBACK(on_location_changed), browser);
    GtkToolItem* toolitem = gtk_tool_item_new();
    gtk_tool_item_set_expand(GTK_TOOL_ITEM(toolitem), TRUE);
    gtk_container_add(GTK_CONTAINER(toolitem), browser->location);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->navibar), toolitem, -1);

    // Search entry
    browser->popup_searchEngines = gtk_ui_manager_get_widget(ui_manager, "/popup_searchEngines");
    g_object_ref(browser->popup_searchEngines);
    menuitem = gtk_separator_menu_item_new();
    gtk_menu_shell_prepend(GTK_MENU_SHELL(browser->popup_searchEngines), menuitem);
    gtk_widget_show(menuitem);
    for(i = 0; i < searchEngines->len; i++)
    {
     CSearchEngine* searchEngine = (CSearchEngine*)g_ptr_array_index(searchEngines, i);
     menuitem = gtk_image_menu_item_new_with_label(searchEngine->sShortName);
     GdkPixbuf* pixbuf = load_web_icon(searchEngine->sIcon, GTK_ICON_SIZE_MENU, menuitem);
     GtkWidget* icon = gtk_image_new_from_pixbuf(pixbuf);
     gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(menuitem), icon);
     g_object_unref(pixbuf);
     gtk_menu_shell_prepend(GTK_MENU_SHELL(browser->popup_searchEngines), menuitem);
     g_object_set_data(G_OBJECT(menuitem), "engine", GINT_TO_POINTER(i));
     g_signal_connect(menuitem, "activate", G_CALLBACK(on_webSearch_engine_activate), browser);
     gtk_widget_show(menuitem);
    }
    browser->webSearch = sexy_icon_entry_new();
    sexy_icon_entry_set_icon_highlight(SEXY_ICON_ENTRY(browser->webSearch)
     , SEXY_ICON_ENTRY_PRIMARY, TRUE);
    // TODO: Make this actively resizable or enlarge to fit contents?
    // FIXME: The interface is somewhat awkward and ought to be rethought
    // TODO: Display "show in context menu" search engines as "completion actions"
    entry_setup_completion(GTK_ENTRY(browser->webSearch));
    sokoke_entry_set_can_undo(GTK_ENTRY(browser->webSearch), TRUE);
    update_searchEngine(config->iSearchEngine, browser);
    g_signal_connect(browser->webSearch, "icon-released"
     , G_CALLBACK(on_webSearch_icon_released), browser);
    g_signal_connect(browser->webSearch, "key-press-event"
     , G_CALLBACK(on_webSearch_key_down), browser);
    g_signal_connect(browser->webSearch, "scroll-event"
     , G_CALLBACK(on_webSearch_scroll), browser);
    g_signal_connect(browser->webSearch, "activate"
     , G_CALLBACK(on_webSearch_activate), browser);
    toolitem = gtk_tool_item_new();
    gtk_container_add(GTK_CONTAINER(toolitem), browser->webSearch);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->navibar), toolitem, -1);
    action = gtk_action_group_get_action(browser->actiongroup, "TabsClosed");
    browser->closedTabs = gtk_action_create_tool_item(action);
    g_signal_connect(browser->closedTabs, "clicked", G_CALLBACK(on_navibar_tabsclosed_clicked), browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->navibar), GTK_TOOL_ITEM(browser->closedTabs), -1);
    sokoke_container_show_children(GTK_CONTAINER(browser->navibar));

    // Bookmarkbar
    browser->bookmarkbar = gtk_toolbar_new();
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(browser->bookmarkbar), GTK_ICON_SIZE_MENU);
    gtk_toolbar_set_style(GTK_TOOLBAR(browser->bookmarkbar), GTK_TOOLBAR_BOTH_HORIZ);
    create_bookmark_menu(bookmarks, browser->menu_bookmarks, browser);
    for(i = 0; i < xbel_folder_get_n_items(bookmarks); i++)
    {
     XBELitem* item = xbel_folder_get_nth_item(bookmarks, i);
     // FIXME: A separator has none of these two
     const gchar* sTitle = xbel_item_get_title(item);
     const gchar* sDesc = xbel_item_get_desc(item);
     switch(xbel_item_get_type(item))
     {
      case XBEL_ITEM_FOLDER:
       toolitem = tool_button_new(sTitle, GTK_STOCK_DIRECTORY, TRUE, TRUE
        , G_CALLBACK(on_bookmark_toolbar_folder_activate), sDesc, browser);
       g_object_set_data(G_OBJECT(toolitem), "xbelitem", item);
       break;
      case XBEL_ITEM_BOOKMARK:
       toolitem = tool_button_new(sTitle, STOCK_BOOKMARK, TRUE, TRUE
        , G_CALLBACK(on_menu_bookmarks_item_activate), sDesc, browser);
       g_object_set_data(G_OBJECT(toolitem), "xbelitem", item);
       break;
      case XBEL_ITEM_SEPARATOR:
       toolitem = gtk_separator_tool_item_new();
       break;
      default:
       g_warning("Unknown item type");
     }
     gtk_toolbar_insert(GTK_TOOLBAR(browser->bookmarkbar), toolitem, -1);
    }
    sokoke_container_show_children(GTK_CONTAINER(browser->bookmarkbar));
    gtk_box_pack_start(GTK_BOX(vbox), browser->bookmarkbar, FALSE, FALSE, 0);

    // Superuser warning
    if(hbox = sokoke_superuser_warning_new())
     gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

    // Create the panels
    GtkWidget* hpaned = gtk_hpaned_new();
    gtk_paned_set_position(GTK_PANED(hpaned), config->iWinPanelPos);
    g_signal_connect(hpaned, "notify::position"
     , GTK_SIGNAL_FUNC(on_panels_notify_position), browser);
    gtk_box_pack_start(GTK_BOX(vbox), hpaned, TRUE, TRUE, 0);
    gtk_widget_show(hpaned);

    browser->panels = gtk_hbox_new(FALSE, 0);
    gtk_paned_pack1(GTK_PANED(hpaned), browser->panels, FALSE, FALSE);
    sokoke_widget_set_visible(browser->panels, config->bPanels);

    // Create the panel toolbar
    GtkWidget* panelbar = gtk_ui_manager_get_widget(ui_manager, "/toolbar_panels");
    gtk_toolbar_set_style(GTK_TOOLBAR(panelbar), GTK_TOOLBAR_BOTH);
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(panelbar), GTK_ICON_SIZE_BUTTON);
    gtk_toolbar_set_orientation(GTK_TOOLBAR(panelbar), GTK_ORIENTATION_VERTICAL); 
    gtk_box_pack_start(GTK_BOX(browser->panels), panelbar, FALSE, FALSE, 0);
    g_object_unref(ui_manager);

    GtkWidget* cbox = gtk_vbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(browser->panels), cbox, TRUE, TRUE, 0);
    gtk_widget_show(cbox);

    // Panels title bar
    GtkWidget* labelbar = gtk_toolbar_new();
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(labelbar), GTK_ICON_SIZE_MENU);
    gtk_toolbar_set_style(GTK_TOOLBAR(labelbar), GTK_TOOLBAR_ICONS);
    toolitem = gtk_tool_item_new();
    gtk_tool_item_set_expand(toolitem, TRUE);
    label = gtk_label_new_with_mnemonic("_Panels");
    gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
    gtk_container_add(GTK_CONTAINER(toolitem), label);
    gtk_container_set_border_width(GTK_CONTAINER(toolitem), 6);
    gtk_toolbar_insert(GTK_TOOLBAR(labelbar), toolitem, -1);
    // TODO: Does 'goto top' actually indicate 'detach'?
    toolitem = tool_button_new(NULL, GTK_STOCK_GOTO_TOP, FALSE, TRUE
     , NULL/*G_CALLBACK(on_panels_button_float_clicked)*/, "Detach panel", browser);
    gtk_toolbar_insert(GTK_TOOLBAR(labelbar), toolitem, -1);
    toolitem = tool_button_new(NULL, GTK_STOCK_CLOSE, FALSE, TRUE
     , G_CALLBACK(on_panels_button_close_clicked), "Close panel", browser);
    gtk_toolbar_insert(GTK_TOOLBAR(labelbar), toolitem, -1);
    gtk_box_pack_start(GTK_BOX(cbox), labelbar, FALSE, FALSE, 0);
    gtk_widget_show_all(labelbar);

    // Notebook, containing all panels
    browser->panels_notebook = gtk_notebook_new();
    gtk_notebook_set_show_border(GTK_NOTEBOOK(browser->panels_notebook), FALSE);
    gtk_notebook_set_show_tabs(GTK_NOTEBOOK(browser->panels_notebook), FALSE);
      GtkWidget* scrolled; GtkWidget* panel; gint iPage;
      // Dummy: This is the "fallback" panel for now
      iPage = gtk_notebook_append_page(GTK_NOTEBOOK(
       browser->panels_notebook), gtk_label_new("empty"), NULL);
      // Pageholder
      #ifdef MIDORI_WEBKIT
      scrolled = gtk_scrolled_window_new(NULL, NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled)
       , GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
      GTK_WIDGET_SET_FLAGS(scrolled, GTK_CAN_FOCUS);
      panel = webkit_page_new();
      #else
      scrolled = gtk_event_box_new();
      panel = webi_new();
      #endif
      browser->panel_pageholder = panel;
      //sokoke_widget_set_transparency(browser->window, 0.7);
      gtk_container_add(GTK_CONTAINER(scrolled), panel);
      document_load_uri(panel, config->sPanelPageholder);
      iPage = gtk_notebook_append_page(GTK_NOTEBOOK(
       browser->panels_notebook), scrolled, NULL);
      action = gtk_action_group_get_action(browser->actiongroup, "PanelPageholder");
      g_object_set_data(G_OBJECT(action), "iPage", GINT_TO_POINTER(iPage));
    GtkWidget* frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
    gtk_container_add(GTK_CONTAINER(frame), browser->panels_notebook);
    gtk_box_pack_start(GTK_BOX(cbox), frame, TRUE, TRUE, 0);
    //gtk_notebook_set_current_page(GTK_NOTEBOOK(browser->panels_notebook), config->iPanelActive);
    gtk_widget_show_all(gtk_widget_get_parent(browser->panels_notebook));

    // Notebook, containing all documents
    browser->documents = gtk_notebook_new();
    gtk_notebook_set_scrollable(GTK_NOTEBOOK(browser->documents), TRUE);
    #if GTK_CHECK_VERSION(2, 10, 0)
    //gtk_notebook_set_group_id(GTK_NOTEBOOK(browser->documents), 0);
    #endif
    g_signal_connect_after(GTK_OBJECT(browser->documents), "switch-page"
     , GTK_SIGNAL_FUNC(on_notebook_switch_page), browser);
    gtk_paned_pack2(GTK_PANED(hpaned), browser->documents, FALSE, FALSE);
    gtk_widget_show(browser->documents);

    // Incremental findbar
    browser->findbox = gtk_toolbar_new();
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(browser->findbox), GTK_ICON_SIZE_MENU);
    gtk_toolbar_set_style(GTK_TOOLBAR(browser->findbox), GTK_TOOLBAR_BOTH_HORIZ);
    toolitem = gtk_tool_item_new();
    gtk_container_set_border_width(GTK_CONTAINER(toolitem), 6);
    gtk_container_add(GTK_CONTAINER(toolitem)
     , gtk_label_new_with_mnemonic("_Inline find:"));
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    browser->findbox_text = sexy_icon_entry_new();
    // TODO: Change that to an error icon if a search fails
    GtkWidget* icon = gtk_image_new_from_stock(GTK_STOCK_FIND, GTK_ICON_SIZE_MENU);
    sexy_icon_entry_set_icon(SEXY_ICON_ENTRY(browser->findbox_text)
     , SEXY_ICON_ENTRY_PRIMARY, GTK_IMAGE(icon));
    sexy_icon_entry_add_clear_button(SEXY_ICON_ENTRY(browser->findbox_text));
    sokoke_entry_set_can_undo(GTK_ENTRY(browser->findbox_text), TRUE);
    g_signal_connect(browser->findbox_text, "activate"
     , GTK_SIGNAL_FUNC(on_action_find_next_activate), browser);
    toolitem = gtk_tool_item_new();
    gtk_container_add(GTK_CONTAINER(toolitem), browser->findbox_text);
    gtk_tool_item_set_expand(GTK_TOOL_ITEM(toolitem), TRUE);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    toolitem = tool_button_new(NULL, GTK_STOCK_GO_BACK, TRUE, TRUE
     , G_CALLBACK(on_action_find_previous_activate), NULL, browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    toolitem = tool_button_new(NULL, GTK_STOCK_GO_FORWARD, TRUE, TRUE
     , G_CALLBACK(on_action_find_next_activate), NULL, browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    browser->findbox_case = gtk_toggle_tool_button_new_from_stock(GTK_STOCK_SPELL_CHECK);
    gtk_tool_button_set_label(GTK_TOOL_BUTTON(browser->findbox_case), "Match case");
    gtk_tool_item_set_is_important(GTK_TOOL_ITEM(browser->findbox_case), TRUE);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), browser->findbox_case, -1);
    toolitem = gtk_separator_tool_item_new();
    gtk_separator_tool_item_set_draw(GTK_SEPARATOR_TOOL_ITEM(toolitem), FALSE);
    gtk_tool_item_set_expand(GTK_TOOL_ITEM(toolitem), TRUE);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    toolitem = tool_button_new(NULL, GTK_STOCK_CLOSE, FALSE, TRUE
     , G_CALLBACK(on_findbox_button_close_clicked), "Close findbar", browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    sokoke_container_show_children(GTK_CONTAINER(browser->findbox));
    gtk_box_pack_start(GTK_BOX(vbox), browser->findbox, FALSE, FALSE, 0);

    // TODO: fix children overlapping statusbar border
    browser->statusbar = gtk_statusbar_new();
    gtk_box_pack_start(GTK_BOX(vbox), browser->statusbar, FALSE, FALSE, 0);
    browser->progress = gtk_progress_bar_new();
    // Setting the progressbar's height to 1 makes it fit in the statusbar
    gtk_widget_set_size_request(browser->progress, -1, 1);
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->progress
     , FALSE, FALSE, 3);

    browser->icon_link_script = gtk_image_new_from_icon_name(
     STOCK_SCRIPT, GTK_ICON_SIZE_MENU);
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->icon_link_script
     , FALSE, FALSE, 0);
    browser->icon_link_window = gtk_image_new_from_icon_name(
     STOCK_WINDOW_NEW, GTK_ICON_SIZE_MENU);
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->icon_link_window
     , FALSE, FALSE, 0);
    browser->icon_security = gtk_image_new();
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->icon_security
     , FALSE, FALSE, 0);
    gtk_widget_show(browser->icon_security);
    browser->icon_newsfeed = gtk_image_new_from_icon_name(STOCK_NEWSFEED
     , GTK_ICON_SIZE_MENU);
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->icon_newsfeed
     , FALSE, FALSE, 0);

    }
    else
    {

    browser->window = oldBrowser->window;
    browser->actiongroup = oldBrowser->actiongroup;
    browser->menubar = oldBrowser->menubar;
    browser->menu_tabsclosed = oldBrowser->menu_tabsclosed;
    browser->menu_bookmarks = oldBrowser->menu_bookmarks;
    browser->menu_window = oldBrowser->menu_window;
    browser->popup_document = oldBrowser->popup_document;
    browser->popup_element = oldBrowser->popup_element;
    browser->throbber = oldBrowser->throbber;
    browser->navibar = oldBrowser->navibar;
    browser->newTab = oldBrowser->newTab;
    browser->location_icon = oldBrowser->location_icon;
    browser->location = oldBrowser->location;
    browser->webSearch = oldBrowser->webSearch;
    browser->closedTabs = oldBrowser->closedTabs;
    browser->popup_searchEngines = oldBrowser->popup_searchEngines;
    browser->bookmarkbar = oldBrowser->bookmarkbar;
    browser->panels = oldBrowser->panels;
    browser->panels_notebook = oldBrowser->panels_notebook;
    browser->panel_pageholder = oldBrowser->panel_pageholder;
    browser->documents = oldBrowser->documents;
    browser->findbox = oldBrowser->findbox;
    browser->statusbar = oldBrowser->statusbar;
    browser->icon_link_script = oldBrowser->icon_link_script;
    browser->icon_link_window = oldBrowser->icon_link_window;
    browser->progress = oldBrowser->progress;
    browser->icon_security = oldBrowser->icon_security;
    browser->icon_newsfeed = oldBrowser->icon_newsfeed;

    }

    // Define some default values
    browser->bMenubar = TRUE;
    browser->bToolbar = TRUE;
    browser->bLocation = TRUE;
    browser->bStatusbar = TRUE;
    browser->bResizable = TRUE;
    browser->sElementUri = NULL;
    browser->iLoadedPercent = -1; // initially "not loading"

    // Add a window menu item
    // TODO: Menu items should be ordered like the notebook tabs
    // TODO: Watch tab reordering in >= gtk 2.10
    browser->document_menu = menu_item_new("about:blank", GTK_STOCK_FILE
     , G_CALLBACK(on_window_menu_item_activate), TRUE, browser);
    gtk_widget_show(browser->document_menu);
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menu_window), browser->document_menu);

    // Create a new page
    GtkWidget* eventbox = gtk_event_box_new();
    gtk_event_box_set_visible_window(GTK_EVENT_BOX(eventbox), FALSE);
    g_signal_connect(GTK_OBJECT(eventbox), "button-release-event"
     , GTK_SIGNAL_FUNC(on_notebook_tab_mouse_up), browser);
    GtkWidget* hbox = gtk_hbox_new(FALSE, 1);
    gtk_container_add(GTK_CONTAINER(eventbox), GTK_WIDGET(hbox));
    browser->document_icon = gtk_image_new_from_stock(GTK_STOCK_FILE, GTK_ICON_SIZE_MENU);
    gtk_box_pack_start(GTK_BOX(hbox), browser->document_icon, FALSE, FALSE, 0);
    browser->document_name = gtk_label_new("about:blank");
    gtk_misc_set_alignment(GTK_MISC(browser->document_name), 0.0, 0.5);
    // TODO: make the tab initially look "unvisited" until it's focused
    // TODO: tabs should shrink when there is not enough space
    // TODO: gtk's tab scrolling is weird?
    gint w, h; sokoke_widget_get_text_size(browser->document_name, "M", &w, &h);
    gtk_widget_set_size_request(GTK_WIDGET(browser->document_name)
     , w * config->iTabSize, -1);
    gtk_label_set_ellipsize(GTK_LABEL(browser->document_name), PANGO_ELLIPSIZE_END);
    gtk_box_pack_start(GTK_BOX(hbox), browser->document_name, FALSE, FALSE, 0);
    browser->document_close = gtk_button_new();
    gtk_button_set_relief(GTK_BUTTON(browser->document_close), GTK_RELIEF_NONE);
    gtk_button_set_focus_on_click(GTK_BUTTON(browser->document_close), FALSE);
    GtkRcStyle* rcstyle = gtk_rc_style_new();
    rcstyle->xthickness = rcstyle->ythickness = 0;
    gtk_widget_modify_style(browser->document_close, rcstyle);
    gtk_rc_style_unref(rcstyle);
    GtkWidget* image = gtk_image_new_from_stock(GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
    gtk_button_set_image(GTK_BUTTON(browser->document_close), image);
    gtk_box_pack_start(GTK_BOX(hbox), browser->document_close, FALSE, FALSE, 0);
    GtkSettings* gtksettings = gtk_settings_get_default(); gint height;
    gtk_icon_size_lookup_for_settings(gtksettings, GTK_ICON_SIZE_BUTTON, 0, &height);
    gtk_widget_set_size_request(browser->document_close, -1, height);
    gtk_widget_show_all(GTK_WIDGET(eventbox));
    sokoke_widget_set_visible(browser->document_close, config->bTabClose);
    g_signal_connect(GTK_OBJECT(browser->document_close), "clicked"
     , GTK_SIGNAL_FUNC(on_notebook_tab_close_clicked), browser);
    #ifdef MIDORI_WEBKIT
    GtkWidget* scrolled = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled)
     , GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    GTK_WIDGET_SET_FLAGS(scrolled, GTK_CAN_FOCUS);
    #else
    GtkWidget* scrolled = gtk_event_box_new();
    #endif
    g_signal_connect(GTK_OBJECT(scrolled), "popup-menu"
     , GTK_SIGNAL_FUNC(on_document_popup), browser);
    g_signal_connect(GTK_OBJECT(scrolled), "scroll-event"
     , GTK_SIGNAL_FUNC(on_document_scroll), browser);
    gint page = gtk_notebook_insert_page(GTK_NOTEBOOK(browser->documents)
     , scrolled, GTK_WIDGET(eventbox)
     , gtk_notebook_get_current_page(GTK_NOTEBOOK(browser->documents)) + 1);
    #if GTK_CHECK_VERSION(2, 10, 0)
    gtk_notebook_set_tab_reorderable(GTK_NOTEBOOK(browser->documents)
     , scrolled, TRUE);
    gtk_notebook_set_tab_detachable(GTK_NOTEBOOK(browser->documents)
     , scrolled, TRUE);
    #endif

    // Create a new document
    #define DOC_CONNECT(__sig, __func) g_signal_connect(\
     G_OBJECT(browser->document), __sig\
     , G_CALLBACK(__func), browser);
    #ifdef MIDORI_WEBI
    browser->document = webi_new();
    webi_set_emit_internal_status(WEBI(browser->document), TRUE); // crap
    gtk_widget_show_all(browser->document); // bug
    DOC_CONNECT("location"             , on_document_location_changed)
    DOC_CONNECT("title"                , on_document_title_changed)
    DOC_CONNECT("load-start"           , on_document_load_started)
    DOC_CONNECT("load-stop"            , on_document_load_finished)
    DOC_CONNECT("status"               , on_document_load_changed)
    DOC_CONNECT("selection"            , on_document_selection_changed)
    DOC_CONNECT("req-js-prompt"        , on_document_request_script_prompt)
    DOC_CONNECT("req-auth-prompt"      , on_document_request_auth_prompt)
    DOC_CONNECT("req-new-window"       , on_document_window_open)
    DOC_CONNECT("set-window-properties", on_document_window_features)
    DOC_CONNECT("set-cookie"           , on_document_set_cookie)
    DOC_CONNECT("mouse-over"           , on_document_link_hover)
    #elif defined(MIDORI_WEBKIT)
    browser->document = GTK_WIDGET(webkit_page_new());
    WebKitFrame* mainframe = webkit_page_get_main_frame(
     WEBKIT_PAGE(browser->document));
    #define FRAME_CONNECT(__sig, __func) g_signal_connect(\
     G_OBJECT(mainframe), __sig\
     , G_CALLBACK(__func), browser);
    /*gchar* sAgentName = g_strconcat(PACKAGE_NAME, "/", PACKAGE_VERSION, NULL);
    webkit_page_set_application_name_for_user_agent(
     PACKAGE_NAME, PACKAGE_VERSION);
    //webkit_page_set_custom_user_agent(sAgentName);
    g_free(sAgentName);*/
    FRAME_CONNECT("title-changed"               , on_frame_title_changed)
    DOC_CONNECT  ("title-changed"               , on_document_title_changed)
    DOC_CONNECT  ("icon-loaded"                 , on_document_icon_changed)
    DOC_CONNECT  ("load-started"                , on_document_load_started)
    DOC_CONNECT  ("load-progress-changed"       , on_document_load_changed)
    DOC_CONNECT  ("load-finished"               , on_document_load_finished)
    FRAME_CONNECT("cleared"                     , on_frame_cleared)
    FRAME_CONNECT("load-done"                   , on_frame_load_done)
    DOC_CONNECT  ("status-bar-text-changed"     , on_document_status_message)
    DOC_CONNECT  ("hovering-over-link"          , on_document_link_hover)
    DOC_CONNECT  ("console-message"        , on_document_console_message)
    DOC_CONNECT  ("script-alert"           , on_document_script_alert)
    DOC_CONNECT  ("script-confirm"         , on_document_script_confirm)
    DOC_CONNECT  ("script-prompt"          , on_document_script_prompt)

    // For now we check for "java-enabled", in case this build has no properties
    if(g_object_class_find_property(G_OBJECT_GET_CLASS(browser->document), "java-enabled"))
    g_object_set(G_OBJECT(browser->document)
     , "loads-images-automatically"      , config->bLoadImagesAutomatically
     , "shrinks-standalone-images-to-fit", config->bShrinkImagesToFit
     , "text-areas-are-resizable"        , config->bResizableTextAreas
     , "java-script-enabled"             , config->bEnableJavaScript
     , "java-enabled"                    , config->bEnableJava
     , "plugins-enabled"                 , config->bEnablePlugins
     , NULL);
    #endif

    DOC_CONNECT  ("button-press-event"     , on_document_mouse_down)
    DOC_CONNECT  ("leave-notify-event"     , on_document_leave)
    DOC_CONNECT  ("destroy"                , on_document_destroy)
    #undef DOC_CONNECT
    g_object_set_data(G_OBJECT(browser->document), "uri", "");
    g_object_set_data(G_OBJECT(browser->document), "title", "about:blank");

    // Eventually pack and display everything
    gtk_container_add(GTK_CONTAINER(scrolled), browser->document);
    sokoke_widget_set_visible(browser->navibar, config->bToolbarNavigation);
    sokoke_widget_set_visible(browser->newTab, config->bToolbarNewTab);
    sokoke_widget_set_visible(browser->webSearch, config->bToolbarWebSearch);
    sokoke_widget_set_visible(browser->closedTabs, config->bToolbarClosedTabs);
    sokoke_widget_set_visible(browser->bookmarkbar, config->bToolbarBookmarks);
    sokoke_widget_set_visible(browser->statusbar, config->bToolbarStatus);
    gtk_widget_show(GTK_WIDGET(scrolled));
    gtk_widget_show(GTK_WIDGET(browser->document));
    if(!config->bOpenTabsInTheBackground)
     gtk_notebook_set_current_page(GTK_NOTEBOOK(browser->documents), page);

    update_browser_actions(browser);
    gtk_widget_show(browser->window);
    gtk_widget_grab_focus(GTK_WIDGET(browser->location));

    return browser;
}
