/* Copyright (C) 2007 Christian Dywan <christian at twotoasts dot de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. */

#include "document.h"

void on_document_location_changed(GtkWidget* document
#ifdef MIDORI_WEBKIT
 , const gchar* sUri
#endif
, CBrowser* browser)
{
    #ifndef MIDORI_WEBKIT
    const gchar* sUri = document_get_uri(document);
    #endif
    // TODO: Ask webkit wether it knows the protocol for "unknown protcol"
    // TODO: This isn't the place for uri scheme handling, is it?
    // TODO: Does this check suffice to indicate an uri?
    // FIXME: sProtocol is leaking, fix that beautifully
    gchar* sProtocol = strtok(g_strdup(sUri), ":");
    if(spawn_protocol_command(sProtocol, sUri))
    {
     //document_stop(document); // TODO: This hack would crash webkit
     return;
    }
    gchar* sNewUri = g_strdup(sUri ? sUri : "");
    g_object_set_data(G_OBJECT(document), "uri", sNewUri);
    if(document == get_current_document(browser))
    {
     gtk_entry_set_text(GTK_ENTRY(browser->location), sNewUri);
     gtk_label_set_text(GTK_LABEL(browser->document_name), sNewUri);
     update_status_message(NULL, browser);
     update_gui_state(browser);
    }
}

void on_document_title_changed(GtkWidget* document
#ifdef MIDORI_WEBKIT
 , const gchar* sTitle, const gchar* sUri
#endif
 , CBrowser* browser)
{
    #ifdef MIDORI_WEBKIT
    // TODO: We emulate location_changed here for now
    // Shouldn't we have separated title_changed and location_changed signals?
    on_document_location_changed(document, sUri, browser);
    #endif
    #ifndef MIDORI_WEBKIT
    const gchar* sTitle = document_get_title(document);
    #endif
    gchar* sNewTitle;
    if(sTitle)
     sNewTitle = g_strdup(sTitle);
    else
    {
     const gchar* sUri = g_object_get_data(G_OBJECT(browser->document), "uri");
     sNewTitle = g_strdup(sUri);
    }
    g_object_set_data(G_OBJECT(document), "title", sNewTitle);
    gtk_label_set_text(GTK_LABEL(browser->document_name), sNewTitle);
    sokoke_widget_set_tooltip_text(gtk_widget_get_parent(
     gtk_widget_get_parent(browser->document_name)), sNewTitle);
    gtk_label_set_text(GTK_LABEL(gtk_bin_get_child(GTK_BIN(
     browser->document_menu))), sNewTitle);
    if(document == get_current_document(browser))
    {
     gchar* sWindowTitle = g_strconcat(sNewTitle, " - ", PACKAGE_NAME, NULL);
     gtk_window_set_title(GTK_WINDOW(browser->window), sWindowTitle);
     g_free(sWindowTitle);
    }
}

#ifdef MIDORI_WEBKIT
void on_frame_title_changed(WebKitFrame* frame
 , const gchar* sTitle, const gchar* sUri, CBrowser* browser)
{
    UNIMPLEMENTED
}

void on_document_icon_changed(GtkWidget* document, WebKitFrame* widget
 , CBrowser* browser)
{
    // TODO: Implement icon updates; currently this isn't ever called anyway
    const gchar* sIcon = NULL;
    UNIMPLEMENTED
    if(sIcon)
    {
     gtk_label_set_text(GTK_LABEL(browser->document_name), "icon");
     gtk_label_set_text(GTK_LABEL(gtk_bin_get_child(GTK_BIN(
      browser->document_menu))), "icon");
     if(document == get_current_document(browser))
     {
      gchar* sWindowTitle = g_strconcat("icon", " - ", PACKAGE_NAME, NULL);
      gtk_window_set_title(GTK_WINDOW(browser->window), sWindowTitle);
      g_free(sWindowTitle);
     }
    }
}
#endif

void on_document_load_started(GtkWidget* document
#ifdef MIDORI_WEBKIT
 , WebKitFrame* widget
#endif
, CBrowser* browser)
{
    browser->iLoadedPercent = 0;
    browser->iLoadedBytes = 0;
    browser->iLoadedBytesMax = 0;
    update_favicon(browser);
    if(document == get_current_document(browser)) update_gui_state(browser);
    update_statusbar_text(browser);
    gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(browser->progress), 0.1);
    gtk_widget_show(browser->progress);
}

void on_document_load_changed(GtkWidget* document
#ifdef MIDORI_WEBKIT
 , gint progress
#else
 , const WebiLoadStatus* status
#endif
 , CBrowser* browser)
{
    #ifndef MIDORI_WEBKIT
    browser->iLoadedBytes = status->received;
    browser->iLoadedBytesMax = status->size;
    #else
    browser->iLoadedBytes = progress;
    browser->iLoadedBytes = 100;
    #endif
    if(document == get_current_document(browser)) update_gui_state(browser);
}

void on_document_load_finished(GtkWidget* document
#ifdef MIDORI_WEBKIT
 , WebKitFrame* widget
#endif
, CBrowser* browser)
{
    browser->iLoadedPercent = -1;
    update_favicon(browser);
    if(document == get_current_document(browser)) update_gui_state(browser);
    update_statusbar_text(browser);
    gtk_widget_hide(browser->progress);
}

#ifdef MIDORI_WEBKIT
void on_frame_cleared(WebKitFrame* frame, CBrowser* browser)
{
    UNIMPLEMENTED
}

void on_frame_load_done(WebKitFrame* frame, gboolean bSuccess
 , CBrowser* browser)
{
    UNIMPLEMENTED
}
#endif

void on_document_status_message(GtkWidget* document, gchar* sText
 , CBrowser* browser)
{
    update_status_message(sText, browser);
}

void on_document_selection_changed(GtkWidget* document, CBrowser* browser)
{
    //g_print("%s\n", "selection_changed");
}

gboolean on_document_console_message(GtkWidget* document
 , const gchar* sMessage, gint iLine, const gchar* sSourceId, CBrowser* browser)
{
    return FALSE;
}

typedef enum {
 SCRIPT_DIALOG_ALERT,
 SCRIPT_DIALOG_CONFIRM,
 SCRIPT_DIALOG_PROMPT
 } SCRIPT_DIALOG;

static gboolean document_script_dialog(GtkWidget* document, gpointer frame
 , const gchar* sMessage, SCRIPT_DIALOG iType, const gchar* sDefaultValue
 , gchar** sValue, CBrowser* browser)
{
    GtkMessageType type; GtkButtonsType buttons; gint iDefaultResponse;
    switch(iType)
    {
     case SCRIPT_DIALOG_ALERT:
      type = GTK_MESSAGE_WARNING;
      buttons = GTK_BUTTONS_CLOSE;
      iDefaultResponse = GTK_RESPONSE_CLOSE; break;
     case SCRIPT_DIALOG_CONFIRM:
      type = GTK_MESSAGE_QUESTION;
      buttons = GTK_BUTTONS_YES_NO;
      iDefaultResponse = GTK_RESPONSE_YES; break;
     case SCRIPT_DIALOG_PROMPT:
      type = GTK_MESSAGE_QUESTION;
      buttons = GTK_BUTTONS_OK_CANCEL;
      iDefaultResponse = GTK_RESPONSE_OK;
    }

    GtkWidget* dialog = gtk_message_dialog_new(GTK_WINDOW(browser->window)
     , GTK_DIALOG_DESTROY_WITH_PARENT, type, buttons, sMessage);
    // FIXME: use the frame's uri
    gchar* sTitle = g_strconcat("JavaScript - "
     , document_get_uri(browser->document), NULL);
    gtk_window_set_title(GTK_WINDOW(dialog), sTitle); g_free(sTitle);

    GtkWidget* entry;
    if(iType == SCRIPT_DIALOG_PROMPT)
    {
     entry = gtk_entry_new();
     gtk_entry_set_text(GTK_ENTRY(entry), sDefaultValue);
     gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), entry);
     gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);
     gtk_widget_show(entry);
    }

    gtk_dialog_set_default_response(GTK_DIALOG(dialog), iDefaultResponse);
    gtk_widget_set_sensitive(document, FALSE);
    gint iResponse = sokoke_dialog_run_modeless(GTK_DIALOG(dialog));
    gtk_widget_set_sensitive(document, TRUE);

    gboolean bReturnValue;
    switch(iResponse)
    {
     case GTK_RESPONSE_YES:
      bReturnValue = TRUE; break;
     case GTK_RESPONSE_NO:
      bReturnValue = FALSE; break;
     case GTK_RESPONSE_OK:
      bReturnValue = TRUE;
      *sValue = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry)));
    }
    gtk_widget_destroy(dialog);
    return bReturnValue;
}

gboolean on_document_script_alert(GtkWidget* document, gpointer frame, const gchar* sMessage
 , CBrowser* browser)
{
    document_script_dialog(document, frame, sMessage
     , SCRIPT_DIALOG_ALERT, NULL, NULL, browser);
    return TRUE;
}

gboolean on_document_script_confirm(GtkWidget* document, gpointer frame
 , const gchar* sMessage, gboolean* bConfirmed, CBrowser* browser)
{
    *bConfirmed = document_script_dialog(document, frame, sMessage
     , SCRIPT_DIALOG_CONFIRM, NULL, NULL, browser);
    return TRUE;
}

gboolean on_document_script_prompt(GtkWidget* document, gpointer frame
 , const gchar* sMessage, const gchar* sDefault, gchar** sValue, CBrowser* browser)
{
    if(!document_script_dialog(document, frame, sMessage
     , SCRIPT_DIALOG_PROMPT, sDefault, sValue, browser))
     *sValue = NULL;
    return TRUE;
}

#ifdef MIDORI_WEBI
void on_document_request_script_prompt(GtkWidget* document
 , WebiPromptArgs* args, CBrowser* browser)
{
    guint iType;
    switch(args->type)
    {
     case WEBI_ALERT:
      iType = SCRIPT_DIALOG_ALERT; break;
     case WEBI_CONFIRM:
      iType = SCRIPT_DIALOG_CONFIRM; break;
     case WEBI_INPUT:
      iType = SCRIPT_DIALOG_PROMPT;
    }

    gchar* sValue;
    gboolean retval = document_script_dialog(document, args->msg, iType
     , args->default_input, sValue, browser);

    if(args->type == WEBI_ALERT || args->type == WEBI_CONFIRM)
     args->out_ok_pressed = retval;
    else
     args->out_input = sValue;
}
#endif

void on_document_request_auth_prompt(GtkWidget* document
#ifdef MIDORI_WEBI
 , WebiAuthArgs* args
#endif
 , CBrowser* browser)
{
    GtkDialog* dialog;
    gchar* title; gchar* msg;
    GtkWidget* hbox;
    GtkSizeGroup* sizegroup; GtkWidget* align; GtkWidget* label;
    GtkWidget* user; GtkWidget* pass;
    gint response;

    #ifdef MIDORI_WEBI
    msg = g_strdup_printf("Authentication required for:\n\n%s", args->realm);
    #endif
    dialog = GTK_DIALOG(gtk_message_dialog_new(GTK_WINDOW(browser->window)
     , GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_OTHER, GTK_BUTTONS_OK_CANCEL
     , msg)); g_free(msg);
    #ifdef MIDORI_WEBI
    title = g_strconcat("Authentication - ", args->url, NULL);
    #endif
    gtk_window_set_title(GTK_WINDOW(dialog), title); g_free(title);
    #if GTK_CHECK_VERSION(2, 10, 0)
    GtkWidget* image = gtk_image_new_from_icon_name(
     GTK_STOCK_DIALOG_AUTHENTICATION, GTK_ICON_SIZE_DIALOG);
    gtk_widget_show(image);
    gtk_message_dialog_set_image(GTK_MESSAGE_DIALOG(dialog), image);
    #endif
    sizegroup = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
    hbox = gtk_hbox_new(FALSE, 6);
    label = gtk_label_new("Username");
    align = gtk_alignment_new(0, 0.5, 0, 0);
    gtk_container_add(GTK_CONTAINER(align), label);
    gtk_size_group_add_widget(sizegroup, align);
    gtk_box_pack_start(GTK_BOX(hbox), align, FALSE, FALSE, 0);
    user = gtk_entry_new();
    gtk_box_pack_start(GTK_BOX(hbox), user, FALSE, FALSE, 0);
    gtk_container_add(GTK_CONTAINER(dialog->vbox), hbox);
    gtk_widget_show_all(hbox);
    hbox = gtk_hbox_new(FALSE, 6);
    label = gtk_label_new("Password");
    align = gtk_alignment_new(0, 0.5, 0, 0);
    gtk_container_add(GTK_CONTAINER(align), label);
    gtk_size_group_add_widget(sizegroup, align);
    gtk_box_pack_start(GTK_BOX(hbox), align, FALSE, FALSE, 0);
    pass = gtk_entry_new();
    gtk_entry_set_visibility(GTK_ENTRY(pass), FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), pass, FALSE, FALSE, 0);
    gtk_container_add(GTK_CONTAINER(dialog->vbox), hbox);
    gtk_widget_show_all(hbox);
    gtk_dialog_set_default_response(dialog, GTK_RESPONSE_CANCEL);

    gtk_widget_set_sensitive(document, FALSE);
    response = sokoke_dialog_run_modeless(dialog);
    gtk_widget_set_sensitive(document, TRUE);

    #ifdef MIDORI_WEBI
    if(response == GTK_RESPONSE_OK)
    {
     args->out_username = g_strdup(gtk_entry_get_text(GTK_ENTRY(user)));
     args->out_password = g_strdup(gtk_entry_get_text(GTK_ENTRY(pass)));
     args->out_ok_pressed = TRUE;
    }
    else
     args->out_ok_pressed = FALSE;
    #endif
    gtk_widget_destroy(GTK_WIDGET(dialog));
}

gboolean on_document_set_cookie(GtkWidget* document
#ifdef MIDORI_WEBI
 , WebiCookie* cookie
#endif
 , CBrowser* browser)
{
    // Whole cookie
    // const gchar* cookie;

    // Cookie name
    // const gchar* name;

    // Cookie value
    // const gchar* value;

    // Cookie comment
    // const gchar* comment;

    // Cookie domain
    // const gchar* domain;

    // Cookie TTL
    // const gchar* maxAge;

    // Cookie path
    // const gchar* path;

    // Is cookie secure?
    // gboolean secure;

    // Cookie-specification version.
    // gint version;

    return TRUE;
}

void on_document_link_hover(GtkWidget* document, const gchar* sTooltip
#ifdef MIDORI_WEBI
 , const gchar* sLabel
#endif
 , const gchar* sUri
#ifdef MIDORI_WEBI
 , const gchar* sTarget
#endif
 , CBrowser* browser)
{
    update_status_message(sUri, browser);
    #ifdef MIDORI_WEBI
    sokoke_widget_set_visible(browser->icon_link_window
     , sTarget && !g_ascii_strncasecmp(sTarget, "_blank", 6));
    // TODO: detect "onclick", "onmousedown", "onmouseup" javascript links
    #endif
    sokoke_widget_set_visible(browser->icon_link_script
     , sUri && g_str_has_prefix(sUri, "javascript:"));
    g_free(browser->sElementUri); browser->sElementUri = g_strdup(sUri);
}

GtkWidget* on_document_window_open(GtkWidget* document, const gchar* sUri
 , CBrowser* browser)
{
    // A window is created
    // TODO: Respect config->iNewPages
    // TODO: Find out if this comes from a script or a click
    // TODO: Block scripted popups, return NULL and show status icon
    CBrowser* newBrowser = browser_new(config->bOpenPopupsInTabs ? browser : NULL);
    return newBrowser->document;
}

void on_document_window_features(GtkWidget* document
#ifdef MIDORI_WEBI
 , WebiWindowProperties* properties
#endif
 , CBrowser* browser)
{
    #ifdef MIDORI_WEBI
    browser->bMenubar = TRUE;
    browser->bToolbar = properties->toolbarsVisible;
    browser->bLocation = properties->toolbarsVisible;
    browser->bStatusbar = properties->statusBarVisible;
    //browser->bScrollbars = properties->scrollbarsVisible;
    browser->bResizable = properties->isResizable;
    #endif
    update_visibility(browser, TRUE);
}

void document_popup(GtkWidget* document, GdkEventButton* event, CBrowser* browser)
{
    gboolean bLink = browser->sElementUri != NULL; // Did we right-click a link?
    gboolean bImage = FALSE; // TODO: Did we right-click an image?
    gchar* sSelection = NULL; // TODO: The possibly selected text

    // Download manager available?
    const gchar* sDownload = g_datalist_get_data(&config->protocols_commands, "download");
    gboolean bDownload = sDownload && strcmp(sDownload, "");

    action_set_visible("LinkTabNew", bLink, browser);
    action_set_visible("LinkTabCurrent", bLink, browser);
    action_set_visible("LinkWindowNew", bLink, browser);

    action_set_visible("LinkSaveAs", bLink, browser);
    action_set_visible("LinkSaveWith", bLink && bDownload, browser);
    action_set_visible("LinkCopy", bLink, browser);
    action_set_visible("LinkBookmarkNew", bLink, browser);

    action_set_visible("SelectionLinksNewTabs", sSelection != NULL, browser);
    action_set_visible("SelectionTextTabNew", sSelection != NULL, browser);
    action_set_visible("SelectionTextTabCurrent", sSelection != NULL, browser);
    action_set_visible("SelectionTextWindowNew", sSelection != NULL, browser);

    action_set_visible("ImageViewTabNew", bImage, browser);
    action_set_visible("ImageViewTabCurrent", bImage, browser);
    action_set_visible("ImageSaveAs", bImage, browser);
    action_set_visible("ImageSaveWith", bImage && bDownload, browser);
    action_set_visible("ImageCopy", bImage, browser);

    action_set_visible("Copy", sSelection != NULL, browser);
    action_set_visible("SelectAll", sSelection != NULL, browser);
    action_set_visible("SelectionSearch", sSelection != NULL, browser);
    action_set_visible("SelectionSearchWith", sSelection != NULL, browser);
    action_set_visible("SelectionSourceView", sSelection != NULL, browser);

    action_set_visible("SourceView", !sSelection, browser);

    if(bLink || bImage)
     sokoke_widget_popup(document, GTK_MENU(browser->popup_element), event);
    else
     sokoke_widget_popup(document, GTK_MENU(browser->popup_document), event);
}

gboolean on_document_mouse_down(GtkWidget* widget, GdkEventButton* event
 , CBrowser* browser)
{
    GdkModifierType state = (GdkModifierType)0;
    gint x, y;
    gdk_window_get_pointer(NULL, &x, &y, &state);
    switch(event->button)
    {
     case 1:
      if(!browser->sElementUri) return FALSE;
      if(state & GDK_SHIFT_MASK)
      {
       // Open link in new window
       CBrowser* curBrowser = browser_new(NULL);
       document_load_uri(curBrowser->document, browser->sElementUri);
       return TRUE;
      }
      else if(state & GDK_MOD1_MASK)
      {
       // Open link in new tab
       CBrowser* curBrowser = browser_new(browser);
       document_load_uri(curBrowser->document, browser->sElementUri);
       return TRUE;
      }
      break;
     case 2:
      if(state & GDK_CONTROL_MASK)
      {
       document_set_text_size(get_current_document(browser), 1);
       return TRUE;
      }
      else
      {
       if(!browser->sElementUri) return FALSE;
       // Open link in new tab
       CBrowser* curBrowser = browser_new(browser);
       document_load_uri(curBrowser->document, browser->sElementUri);
       return TRUE;
      }
      break;
     case 3:
      document_popup(get_current_document(browser), event, browser);
      return TRUE;
    }
    return FALSE;
}

void on_document_popup(GtkWidget* document, CBrowser* browser)
{
    document_popup(get_current_document(browser), NULL, browser);
}

gboolean on_document_scroll(GtkWidget* document, GdkEventScroll* event
 , CBrowser* browser)
{
    GdkModifierType state = (GdkModifierType)0;
    gint x, y;
    gdk_window_get_pointer(NULL, &x, &y, &state);
    if(state & GDK_CONTROL_MASK)
    {
     const gfloat size = document_get_text_size(get_current_document(browser));
     if(event->direction == GDK_SCROLL_DOWN)
      document_set_text_size(get_current_document(browser), size + 0.1);
     else if(event->direction == GDK_SCROLL_UP)
      document_set_text_size(get_current_document(browser), size - 0.1);
     return TRUE;
    }
    else
     return FALSE;
}

void on_document_destroy(GtkWidget* widget, CBrowser* browser)
{
    // Update browser list, free memory and possibly quit
    browser_count = GUINT_TO_POINTER(GPOINTER_TO_UINT(browser_count) - 1);
    GList* tmp_list = g_list_find(browser_list, browser);
    browser_list = g_list_delete_link(browser_list, tmp_list); /*remove?*/
    g_free(browser->sElementUri);
    g_free(browser->sStatusMessage);
    if(!GPOINTER_TO_INT(browser_count))
    {
     g_object_unref(browser->actiongroup);
     g_object_unref(browser->popup_document);
     g_object_unref(browser->popup_element);
     gtk_main_quit();
    }
}

// Document actions begin here

#define DOCUMENT_IS_SENSITIVE(rv) \
 if(!GTK_WIDGET_SENSITIVE(document)) return rv

void document_load_uri(GtkWidget* document, const gchar* sUri)
{
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_load_url(WEBI(document), sUri);
    #elif defined(MIDORI_WEBKIT)
    // TODO: The block below is just for testing :D
    if(sUri && g_str_has_prefix(sUri, "res://"))
    {
     webkit_page_load_html_string(WEBKIT_PAGE(document)
      , "<html><head><title>" PACKAGE_STRING "</title></head>"
        "<body>test</body></html>"
      , "about:about");
     return;
    }
    webkit_page_open(WEBKIT_PAGE(document), (gchar*)sUri);
    #endif
}

void document_go_back(GtkWidget* document)
{
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_go_back(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    webkit_page_go_backward(WEBKIT_PAGE(document));
    #endif
}

void document_go_forward(GtkWidget* document)
{
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_go_forward(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    webkit_page_go_forward(WEBKIT_PAGE(document));
    #endif
}

void document_refresh(GtkWidget* document, gboolean bSkipCache)
{
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_refresh(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    webkit_page_reload(WEBKIT_PAGE(document));
    #endif
}

void document_stop(GtkWidget* document)
{
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_stop_load(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    webkit_page_stop_loading(WEBKIT_PAGE(document));
    #endif
}

gboolean document_can_go_back(GtkWidget* document)
{
    g_return_val_if_fail(GTK_IS_WIDGET(document), FALSE);
    #ifdef MIDORI_WEBI
    return webi_can_go_back(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    return webkit_page_can_go_backward(WEBKIT_PAGE(document));
    #endif
}

gboolean document_can_go_forward(GtkWidget* document)
{
    g_return_val_if_fail(GTK_IS_WIDGET(document), FALSE);
    #ifdef MIDORI_WEBI
    return webi_can_go_forward(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    return webkit_page_can_go_forward(WEBKIT_PAGE(document));
    #endif
}

G_CONST_RETURN gchar* document_get_uri(GtkWidget* document)
{
    g_return_val_if_fail(GTK_IS_WIDGET(document), NULL);
    #ifdef MIDORI_WEBI
    return webi_get_location(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    WebKitFrame* frame = webkit_page_get_main_frame(WEBKIT_PAGE(document));
    webkit_frame_get_location(frame);
    #endif
}

G_CONST_RETURN gchar* document_get_title(GtkWidget* document)
{
    g_return_val_if_fail(GTK_IS_WIDGET(document), NULL);
    #ifdef MIDORI_WEBI
    return webi_get_title(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    WebKitFrame* frame = webkit_page_get_main_frame(WEBKIT_PAGE(document));
    webkit_frame_get_title(frame);
    #endif
}

// FIXME: This better be document_find(document, text, forward, case, wrap)
gboolean document_find(GtkWidget* document, const gchar* sText
 , gboolean bCase, gboolean bForward)
{
    DOCUMENT_IS_SENSITIVE(FALSE);
    #ifdef MIDORI_WEBI
    return webi_find(WEBI(document), sText, bCase, bForward);
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED return FALSE; /*webkit_page_search_for(WEBKIT_PAGE(document)
     , sText, bForward, bCase, TRUE);*/
    #endif
}

gboolean document_can_increase_text_size(GtkWidget* document)
{
    g_return_val_if_fail(GTK_IS_WIDGET(document), FALSE);
    DOCUMENT_IS_SENSITIVE(FALSE);
    #ifdef MIDORI_WEBI
    return TRUE;
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED return FALSE; //webkit_page_can_increase_text_size(WEBKIT_PAGE(document));
    #endif
}

gboolean document_can_decrease_text_size(GtkWidget* document)
{
    g_return_val_if_fail(GTK_IS_WIDGET(document), FALSE);
    DOCUMENT_IS_SENSITIVE(FALSE);
    #ifdef MIDORI_WEBI
    return TRUE;
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED return FALSE; //webkit_page_can_decrease_text_size(WEBKIT_PAGE(document));
    #endif
}

void document_set_text_size(GtkWidget* document, gfloat iSize)
{
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_set_text_multiplier(WEBI(document), iSize);
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED //webkit_page_set_text_size_multiplier(WEBKIT_PAGE(document), iSize);
    #endif
}

gfloat document_get_text_size(GtkWidget* document)
{
    g_return_val_if_fail(GTK_IS_WIDGET(document), 1);
    #ifdef MIDORI_WEBI
    return webi_get_text_multiplier(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED return 1;//webkit_page_get_text_size_multiplier(WEBKIT_PAGE(document));
    #endif
}

gboolean on_document_leave(GtkWidget* document, GdkEventCrossing* event, CBrowser* browser)
{
    update_status_message(NULL, browser);
}

void document_close(GtkWidget* document, CBrowser* browser)
{
    // TODO: Delete old items if > 10 items
    const gchar* sUri = g_object_get_data(G_OBJECT(document), "uri");
    const gchar* sTitle = g_object_get_data(G_OBJECT(document), "title");
    GtkWidget* menuitem = menu_item_new(sTitle, GTK_STOCK_FILE
     , G_CALLBACK(on_menu_tabsclosed_item_activate), TRUE, browser);
    sokoke_widget_set_tooltip_text(menuitem, sUri);
    g_object_set_data(G_OBJECT(menuitem), "uri", (gchar*)sUri);
    gtk_menu_shell_prepend(GTK_MENU_SHELL(browser->menu_tabsclosed), menuitem);
    gtk_widget_show(menuitem);
    gtk_widget_destroy(browser->document_menu);
    gtk_notebook_remove_page(GTK_NOTEBOOK(browser->documents)
     , get_document_index(document, browser));
    update_browser_actions(browser);
}

#undef DOCUMENT_IS_SENSITIVE
