/* Copyright (C) 2007 Christian Dywan <christian at twotoasts dot de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. */

#include "helpers.h"

GtkIconTheme* get_icon_theme(GtkWidget* widget)
{
    return gtk_icon_theme_get_for_screen(gtk_widget_get_screen(widget));
}

GtkWidget* menu_item_new(const gchar* sText, const gchar* sIcon
 , GCallback signal, gboolean bSensitive, gpointer userdata)
{
    // FIXME: Stock id and icon name aren't handled equally.
    GtkWidget* menuitem;
    if(sText)
    {
     menuitem = gtk_image_menu_item_new_with_mnemonic(sText);
     if(sIcon)
     {
      GtkWidget* icon = gtk_image_new_from_stock(sIcon, GTK_ICON_SIZE_MENU);
      if(gtk_image_get_storage_type(GTK_IMAGE(icon)) == GTK_IMAGE_EMPTY)
       icon = gtk_image_new_from_icon_name(sIcon, GTK_ICON_SIZE_MENU);
      if(gtk_image_get_storage_type(GTK_IMAGE(icon)) != GTK_IMAGE_EMPTY)
       gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(menuitem), icon);
      else
       g_print("Note: The icon %s is not available.", sIcon);
     }
    }
    else
     menuitem = gtk_image_menu_item_new_from_stock(sIcon, NULL);
    if(signal) g_signal_connect(menuitem, "activate", signal, userdata);
    gtk_widget_set_sensitive(GTK_WIDGET(menuitem), bSensitive && signal);
    return menuitem;
}

GtkToolItem* tool_button_new(const gchar* sText, const gchar* sIcon
 , gboolean bImportant, gboolean bSensitive, GCallback signal
 , const gchar* sTooltip, gpointer userdata)
{
    // FIXME: Flawed code. Unused widget and stock versus icon name.
    GtkToolItem* toolitem;
    if(sText)
    {
     toolitem = gtk_tool_button_new(NULL, sText);
     GtkWidget* icon = gtk_image_new_from_icon_name(sIcon, GTK_ICON_SIZE_MENU);
     GtkIconTheme* icon_theme = get_icon_theme(GTK_WIDGET(toolitem));
     if(gtk_icon_theme_has_icon(icon_theme, sIcon))
      gtk_tool_button_set_icon_name(GTK_TOOL_BUTTON(toolitem), sIcon);
    }
    else
     toolitem = gtk_tool_button_new_from_stock(sIcon);
    if(bImportant)
     gtk_tool_item_set_is_important(toolitem, TRUE);
    if(signal) g_signal_connect(toolitem, "clicked", signal, userdata);
    gtk_widget_set_sensitive(GTK_WIDGET(toolitem), bSensitive && signal);
    if(sTooltip) sokoke_tool_item_set_tooltip_text(toolitem, sTooltip);
    return toolitem;
}

GtkWidget* check_menu_item_new(const gchar* sText
 , const gchar* sPath, const gchar* sKey, GdkModifierType accel_mods
 , GCallback signal, gboolean bSensitive, gboolean bActive, CBrowser* browser)
{
    GtkWidget* menuitem = gtk_check_menu_item_new_with_mnemonic(sText);
    gtk_widget_set_sensitive(menuitem, bSensitive && signal != NULL);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menuitem), bActive);
    if(signal) g_signal_connect(menuitem, "activate", signal, browser);
    sokoke_menu_item_set_accel(GTK_MENU_ITEM(menuitem), sPath, sKey, accel_mods);
    return menuitem;
}

GtkWidget* radio_button_new(GtkRadioButton* radio_button, const gchar* sLabel)
{
    return gtk_radio_button_new_with_mnemonic_from_widget(radio_button, sLabel);
}

void show_error(const gchar* sText, const gchar* sText2, CBrowser* browser)
{
    GtkWidget* dialog = gtk_message_dialog_new(
     browser ? GTK_WINDOW(browser->window) : NULL
      , 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, sText);
    if(sText2)
     gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog), sText2);
    gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_destroy(dialog);
}

gboolean spawn_protocol_command(const gchar* sProtocol, const gchar* sRes)
{
    const gchar* sCommand = g_datalist_get_data(&config->protocols_commands, sProtocol);
    if(!sCommand) return FALSE;

    // Create an argument vector
    gchar* sUriEscaped = g_shell_quote(sRes);
    gchar* sCommandReady;
    if(strstr(sCommand, "%s"))
     sCommandReady = g_strdup_printf(sCommand, sUriEscaped);
    else
     sCommandReady = g_strconcat(sCommand, " ", sUriEscaped, NULL);
    gchar** argv; GError* error = NULL;
    if(!g_shell_parse_argv(sCommandReady, NULL, &argv, &error))
    {
     // TODO: Should we have a more specific message?
     show_error("Could not run external program.", error->message, NULL);
     g_error_free(error);
     g_free(sCommandReady); g_free(sUriEscaped);
     return FALSE;
    }

    // Try to run the command
    error = NULL;
    gboolean bSuccess = g_spawn_async(NULL, argv, NULL
     , (GSpawnFlags)G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD
     , NULL, NULL, NULL, &error);
    g_strfreev(argv);

    if(!bSuccess)
    {
     // TODO: Should we have a more specific message?
     show_error("Could not run external program.", error->message, NULL);
     g_error_free(error);
    }
    g_free(sCommandReady); g_free(sUriEscaped);
    return TRUE;
}

GdkPixbuf* load_web_icon(const gchar* sIcon, GtkIconSize size, GtkWidget* widget)
{
    // TODO: We want to allow http as well, maybe also base64?
    // FIXME: If we have an icon name it is not theme safe
    GdkPixbuf* pixbuf = NULL; const gchar* sIconReady;
    if(sIcon && g_str_has_prefix(sIcon, "file://"))
     sIconReady = &sIcon[7];
    else
     sIconReady = sIcon;
    if(!(pixbuf = gdk_pixbuf_new_from_file_at_size(sIconReady, 16, 16, NULL))
     && !(pixbuf = gtk_widget_render_icon(widget, sIconReady, size, NULL)))
     pixbuf = gtk_widget_render_icon(widget, GTK_STOCK_FILE, size, NULL);
    return pixbuf;
}

void entry_setup_completion(GtkEntry* entry)
{
    /* TODO: The current behavior works only with the beginning of strings
             But we want to match "localhost" with "loc" and "hos" */
    GtkEntryCompletion* completion = gtk_entry_completion_new();
    gtk_entry_completion_set_model(completion
     , GTK_TREE_MODEL(gtk_list_store_new(1, G_TYPE_STRING)));
    gtk_entry_completion_set_text_column(completion, 0);
    gtk_entry_completion_set_minimum_key_length(completion, 3);
    gtk_entry_set_completion(entry, completion);
}

void entry_completion_append(GtkEntry* entry, const gchar* sText)
{
    GtkEntryCompletion* completion = gtk_entry_get_completion(entry);
    GtkTreeModel* completion_store = gtk_entry_completion_get_model(completion);
    GtkTreeIter iter;
    gtk_list_store_insert(GTK_LIST_STORE(completion_store), &iter, 0);
    gtk_list_store_set(GTK_LIST_STORE(completion_store), &iter, 0, sText, -1);
}

GtkWidget* get_current_document(CBrowser* browser)
{
    gint page = gtk_notebook_get_current_page(GTK_NOTEBOOK(browser->documents));
    GtkWidget* scrolled = gtk_notebook_get_nth_page(GTK_NOTEBOOK(browser->documents), page);
    return gtk_bin_get_child(GTK_BIN(scrolled));
}

gint get_document_index(GtkWidget* document, CBrowser* browser)
{
    GtkWidget* scrolled = gtk_widget_get_parent(document);
    return gtk_notebook_page_num(GTK_NOTEBOOK(browser->documents), scrolled);
}

void update_favicon(CBrowser* browser)
{
    if(browser->iLoadedPercent == -1)
    {
     if(0) //browser->favicon // Has favicon?
     {
      // TODO: use custom icon
      // gtk_image_set_from_file(GTK_IMAGE(browser->icon_page), "image");
     }
     else if(0) // Known mime-type?
     {
      // TODO: Retrieve mime type and load icon; don't forget ftp listings
     }
     else
      gtk_image_set_from_stock(GTK_IMAGE(browser->document_icon), GTK_STOCK_FILE
       , GTK_ICON_SIZE_MENU);
    }
    else
    {
     gtk_image_set_from_stock(GTK_IMAGE(browser->document_icon), GTK_STOCK_EXECUTE
       , GTK_ICON_SIZE_MENU);
    }
}

void update_security(CBrowser* browser)
{
    const gchar* sUri = g_object_get_data(G_OBJECT(browser->document), "uri");
    // TODO: This check is bogus, until webkit tells us how secure a page is
    if(g_str_has_prefix(sUri, "https://"))
    {
     // TODO: highlighted entry indicates security, find an alternative
     gtk_widget_modify_base(browser->location, GTK_STATE_NORMAL
      , &browser->location->style->base[GTK_STATE_SELECTED]);
     gtk_widget_modify_text(browser->location, GTK_STATE_NORMAL
      , &browser->location->style->text[GTK_STATE_SELECTED]);
     gtk_image_set_from_stock(GTK_IMAGE(browser->icon_security)
      , GTK_STOCK_DIALOG_AUTHENTICATION, GTK_ICON_SIZE_MENU);
    }
    else
    {
     gtk_widget_modify_base(browser->location, GTK_STATE_NORMAL
      , NULL);
     gtk_widget_modify_text(browser->location, GTK_STATE_NORMAL
      , NULL);
     gtk_image_set_from_stock(GTK_IMAGE(browser->icon_security)
      , GTK_STOCK_INFO, GTK_ICON_SIZE_MENU);
    }
}

void update_visibility(CBrowser* browser, gboolean visibility)
{
    if(gtk_notebook_get_n_pages(GTK_NOTEBOOK(browser->documents)) > 1
     || 1) return;
    // TODO: Check if this is a scripted window, else just return

    // SHOULD SCRIPTS BE ABLE TO HIDE WINDOWS AT ALL?
    if(0 && !visibility)
    {
     gtk_widget_hide(browser->window);
     return;
    }
    else if(!visibility)
     g_print("Window was not hidden.\n");

    sokoke_widget_set_visible(browser->menubar, browser->bMenubar);
    sokoke_widget_set_visible(browser->navibar, browser->bToolbar);
    sokoke_widget_set_visible(browser->location, browser->bLocation);
    sokoke_widget_set_visible(browser->webSearch, browser->bLocation);
    sokoke_widget_set_visible(browser->statusbar, browser->bStatusbar);
    gtk_window_set_resizable(GTK_WINDOW(browser->window), browser->bResizable);
}

void action_set_active(const gchar* sAction, gboolean bActive
 , CBrowser* browser)
{
    // This shortcut toggles activity state by an action name
    GtkAction* action = gtk_action_group_get_action(browser->actiongroup, sAction);
    g_return_if_fail(GTK_IS_ACTION(action));
    gtk_toggle_action_set_active(GTK_TOGGLE_ACTION(action), bActive);
}

void action_set_sensitive(const gchar* sAction, gboolean bSensitive
 , CBrowser* browser)
{
    // This shortcut toggles sensitivity by an action name
    GtkAction* action = gtk_action_group_get_action(browser->actiongroup, sAction);
    g_return_if_fail(GTK_IS_ACTION(action));
    gtk_action_set_sensitive(action, bSensitive);
}

void action_set_visible(const gchar* sAction, gboolean bVisible
 , CBrowser* browser)
{
    // This shortcut toggles visibility by an action name
    GtkAction* action = gtk_action_group_get_action(browser->actiongroup, sAction);
    g_return_if_fail(GTK_IS_ACTION(action));
    gtk_action_set_visible(action, bVisible);
}

void update_statusbar_text(CBrowser* browser)
{
    if(browser->sStatusMessage)
    {
     gtk_statusbar_pop(GTK_STATUSBAR(browser->statusbar), 1);
     gtk_statusbar_push(GTK_STATUSBAR(browser->statusbar), 1
      , browser->sStatusMessage);
    }
    else
    {
     gchar message[56];
     if(browser->iLoadedPercent)
      g_snprintf(message, 55, "d%% loaded, %d/%d bytes"
       , browser->iLoadedPercent, browser->iLoadedBytes, browser->iLoadedBytesMax);
     else if(browser->iLoadedBytes)
      g_snprintf(message, 55, "%d bytes", browser->iLoadedBytes);
     else
      g_sprintf(message, " ");
     gtk_progress_bar_set_text(GTK_PROGRESS_BAR(browser->progress), message);
    }
}

void update_gui_state(CBrowser* browser)
{
    GtkWidget* document = get_current_document(browser);
    action_set_sensitive("ZoomIn", document_can_increase_text_size(document), browser);
    action_set_sensitive("ZoomOut", document_can_decrease_text_size(document), browser);
    action_set_sensitive("ZoomNormal", document_get_text_size(document) != 1, browser);
    action_set_sensitive("Back", document_can_go_back(document), browser);
    action_set_sensitive("Forward", document_can_go_forward(document), browser);
    action_set_sensitive("Refresh", browser->iLoadedPercent == -1, browser);
    action_set_sensitive("Stop", browser->iLoadedPercent != -1, browser);

    GtkAction* action = gtk_action_group_get_action(browser->actiongroup, "RefreshStop");
    if(browser->iLoadedPercent == -1)
    {
     thunar_throbber_set_animated(THUNAR_THROBBER(browser->throbber), FALSE);
     g_object_set(action, "stock-id", GTK_STOCK_REFRESH, NULL);
     g_object_set(action, "tooltip", "Refresh the current page", NULL);
    }
    else
    {
     thunar_throbber_set_animated(THUNAR_THROBBER(browser->throbber), TRUE);
     g_object_set(action, "stock-id", GTK_STOCK_STOP, NULL);
     g_object_set(action, "tooltip", "Stop loading the current page", NULL);
    }

    gtk_image_set_from_stock(GTK_IMAGE(browser->location_icon), GTK_STOCK_FILE
     , GTK_ICON_SIZE_MENU);

    if(browser->iLoadedBytesMax < 1) // Skip null and negative values
    {
     gtk_progress_bar_pulse(GTK_PROGRESS_BAR(browser->progress));
     update_statusbar_text(browser);
    }
    else
    {
     if(browser->iLoadedBytes > browser->iLoadedBytesMax)
      browser->iLoadedPercent = 100;
     else
      browser->iLoadedPercent = (browser->iLoadedBytes * 100) / browser->iLoadedBytesMax;
     gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(browser->progress)
      , browser->iLoadedPercent / 100.0);
     update_statusbar_text(browser);
    }
}

void update_feeds(CBrowser* browser)
{
    // TODO: Look for available feeds, requires dom access
}

void update_search_engines(CBrowser* browser)
{
    // TODO: Look for available search engines, requires dom access
}

void update_status_message(const gchar* message, CBrowser* browser)
{
    GString* string = g_string_new(message);
    browser->sStatusMessage = g_string_free(string, FALSE);
    update_statusbar_text(browser);
}

void update_browser_actions(CBrowser* browser)
{
    gboolean bActive = gtk_notebook_get_n_pages(GTK_NOTEBOOK(browser->documents)) > 1;
    gtk_notebook_set_show_tabs(GTK_NOTEBOOK(browser->documents), bActive);
    action_set_sensitive("TabClose", bActive, browser);
    GList* children = gtk_container_get_children(
     GTK_CONTAINER(browser->menu_tabsclosed));
    action_set_sensitive("UndoTabClose", g_list_length(children) > 3, browser);
    action_set_sensitive("TabsClosed", g_list_length(children) > 3, browser);
    g_list_free(children);
}

gchar* magic_uri(const gchar* sUri, gboolean bSearch)
{
    // Add file:// if we have a local path
    if(g_path_is_absolute(sUri)) return g_strconcat("file://", sUri, NULL);
    // Do we need to add a protocol?
    if(!strstr(sUri, "://"))
    {
     // Do we have a domain, ip address or localhost?
     if(strstr(sUri, ".") != NULL || !strcmp(sUri, "localhost"))
      return g_strconcat("http://", sUri, NULL);
     // We don't want to search? So return early.
     if(!bSearch) return g_strdup(sUri);
     gchar sSearch[256];
     const gchar* sSearchUrl = NULL;
     // Do we have a keyword and a string?
     gchar** sParts = g_strsplit(sUri, " ", 2);
     if(sParts[0] && sParts[1])
     {
      guint i; for(i = 0; i < searchEngines->len; i++)
      {
       CSearchEngine* searchEngine = (CSearchEngine*)g_ptr_array_index(searchEngines, i);
       if(!strcmp(searchEngine->sKeyword, sParts[0]))
        sSearchUrl = searchEngine->sUrl;
      }
      if(sSearchUrl != NULL)
       g_snprintf(sSearch, 255, sSearchUrl, sParts[1]);
     }
     //g_strfreev(sParts);
     // We only have a word or there is no matching keyowrd, so search for it
     if(sSearchUrl == NULL)
      g_snprintf(sSearch, 255, config->sLocationSearch, sUri);
     return g_strdup(sSearch);
    }
    return g_strdup(sUri);
}

gchar* get_default_font(void)
{
    GtkSettings* gtksettings = gtk_settings_get_default(); gchar* sDefaultFont;
    g_object_get(gtksettings, "gtk-font-name", &sDefaultFont, NULL);
    return sDefaultFont;
}

GtkToolbarStyle config_to_toolbarstyle(guint iToolbarStyle)
{
    switch(iToolbarStyle)
    {
     case CONFIG_TOOLBAR_ICONS:      return GTK_TOOLBAR_ICONS;
     case CONFIG_TOOLBAR_TEXT:       return GTK_TOOLBAR_TEXT;
     case CONFIG_TOOLBAR_BOTH:       return GTK_TOOLBAR_BOTH;
     case CONFIG_TOOLBAR_BOTH_HORIZ: return GTK_TOOLBAR_BOTH_HORIZ;
    }
    GtkSettings* gtksettings = gtk_settings_get_default();
    g_object_get(gtksettings, "gtk-toolbar-style", &iToolbarStyle, NULL);
    return iToolbarStyle;
}

GtkToolbarStyle config_to_toolbariconsize(gboolean bToolbarSmall)
{
    return bToolbarSmall ? GTK_ICON_SIZE_SMALL_TOOLBAR
     : GTK_ICON_SIZE_LARGE_TOOLBAR;
}
