/*
 * Copyright (c) 2005-2006 Benedikt Meurer <benny@xfce.org>
 * Copyright (c) 2007 Christian Dywan <christian at twotoasts dot de>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __THUNAR_THROBBER_H__
#define __THUNAR_THROBBER_H__

#include <gtk/gtk.h>

enum
{
  PROP_0,
  PROP_ANIMATED,
};

G_BEGIN_DECLS;

struct _ThunarThrobberClass
{
  GtkWidgetClass __parent__;
};

struct _ThunarThrobber
{
  GtkWidget __parent__;

  GdkPixbuf *icon;

  gboolean   animated;
  gint       index;
  gint       timer_id;
};

static GObjectClass *thunar_throbber_parent_class;

typedef struct _ThunarThrobberClass ThunarThrobberClass;
typedef struct _ThunarThrobber      ThunarThrobber;

#define THUNAR_TYPE_THROBBER            (thunar_throbber_get_type ())
#define THUNAR_THROBBER(obj)            (G_TYPE_CHECK_INSTANCE_CAST ((obj), THUNAR_TYPE_THROBBER, ThunarThrobber))
#define THUNAR_THROBBER_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), THUNAR_TYPE_THROBBER, ThunarThrobberClass))
#define THUNAR_IS_THROBBER(obj)         (G_TYPE_CHECK_INSTANCE_TYPE ((obj), THUNAR_TYPE_THROBBER))
#define THUNAR_IS_THROBBER_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), THUNAR_TYPE_THROBBER))
#define THUNAR_THROBBER_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), THUNAR_TYPE_THROBBER, ThunarThrobberClass))

GType      thunar_throbber_get_type     (void) G_GNUC_CONST;

GtkWidget *thunar_throbber_new          (void) G_GNUC_MALLOC;

gboolean   thunar_throbber_get_animated (const ThunarThrobber *throbber);
void       thunar_throbber_set_animated (ThunarThrobber       *throbber,
                                         gboolean              animated);

static void     thunar_throbber_class_init    (ThunarThrobberClass  *klass);
static void     thunar_throbber_init          (ThunarThrobber       *throbber);
static void     thunar_throbber_dispose       (GObject              *object);
static void     thunar_throbber_get_property  (GObject              *object,
                                               guint                 prop_id,
                                               GValue               *value,
                                               GParamSpec           *pspec);
static void     thunar_throbber_set_property  (GObject              *object,
                                               guint                 prop_id,
                                               const GValue         *value,
                                               GParamSpec           *pspec);
static void     thunar_throbber_realize       (GtkWidget            *widget);
static void     thunar_throbber_unrealize     (GtkWidget            *widget);
static void     thunar_throbber_size_request  (GtkWidget            *widget,
                                               GtkRequisition       *requisition);
static gboolean thunar_throbber_expose_event  (GtkWidget            *widget,
                                               GdkEventExpose       *event);
static gboolean thunar_throbber_timer         (gpointer              user_data);
static void     thunar_throbber_timer_destroy (gpointer              user_data);

G_END_DECLS;

#endif /* !__THUNAR_THROBBER_H__ */
