/* Copyright (C) 2007 Christian Dywan <christian at twotoasts dot de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. */

#include "webSearch.h"

void update_searchEngine(guint iSearchEngine, CBrowser* browser)
{
    // Display a default icon in case we have no engines
    if(!searchEngines->len)
     sexy_icon_entry_set_icon(SEXY_ICON_ENTRY(browser->webSearch), SEXY_ICON_ENTRY_PRIMARY
     , GTK_IMAGE(gtk_image_new_from_stock(GTK_STOCK_FIND, GTK_ICON_SIZE_MENU)));
    // Change the icon and default text according to the chosen engine
    else if(iSearchEngine < searchEngines->len)
    {
     CSearchEngine* searchEngine = (CSearchEngine*)g_ptr_array_index
      (searchEngines, iSearchEngine);
     GdkPixbuf* pixbuf = load_web_icon(searchEngine->sIcon
      , GTK_ICON_SIZE_MENU, browser->navibar);
     sexy_icon_entry_set_icon(SEXY_ICON_ENTRY(browser->webSearch)
     , SEXY_ICON_ENTRY_PRIMARY, GTK_IMAGE(gtk_image_new_from_pixbuf(pixbuf)));
     g_object_unref(pixbuf);
     sokoke_entry_set_default_text(GTK_ENTRY(browser->webSearch), searchEngine->sShortName);
     config->iSearchEngine = iSearchEngine;
    }
}

void on_webSearch_icon_released(GtkWidget* widget, SexyIconEntryPosition* pos
 , gint button, CBrowser* browser)
{
    sokoke_widget_popup(widget, GTK_MENU(browser->popup_searchEngines), NULL);
}

void on_webSearch_engine_activate(GtkWidget* widget, CBrowser* browser)
{
    guint iSearchEngine = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(widget), "engine"));
    update_searchEngine(iSearchEngine, browser);
}

static void on_webSearch_engines_render_icon(GtkTreeViewColumn* column
 , GtkCellRenderer* renderer, GtkTreeModel* model, GtkTreeIter* iter
 , GtkWidget* treeview)
{
    gchar* sIcon;
    gtk_tree_model_get(model, iter, ENGINES_COL_ICON, &sIcon, -1);

    // TODO: Would it be better, to not do this on every redraw?
    if(sIcon)
    {
     GtkIconTheme* icon_theme = get_icon_theme(treeview);
     // TODO: Is it good to just display nothing if there is no icon?
     GdkPixbuf* icon = load_web_icon(sIcon, GTK_ICON_SIZE_DND, treeview);
     g_object_set(renderer, "pixbuf", icon, NULL);
     if(icon) g_object_unref(icon);
    }
    else
     g_object_set(renderer, "pixbuf", NULL, NULL);
    g_free(sIcon);
}

static void on_webSearch_engines_render_text(GtkTreeViewColumn* column
 , GtkCellRenderer* renderer, GtkTreeModel* model, GtkTreeIter* iter
 , GtkWidget* treeview)
{
    gchar* sName; gchar* sDescription;
    gtk_tree_model_get(model, iter, ENGINES_COL_NAME, &sName
     , ENGINES_COL_DESCRIPTION, &sDescription, -1);
    gchar* sMarkup = g_strdup_printf("<b>%s</b>\n%s", sName, sDescription);
    g_object_set(renderer, "markup", sMarkup, NULL);
    g_free(sName); g_free(sDescription); g_free(sMarkup);
}

static GtkWidget* mini_button_new_from_stock(const gchar* sStockId)
{
    GtkWidget* icon = gtk_image_new_from_stock(sStockId, GTK_ICON_SIZE_BUTTON);
    GtkWidget* button = gtk_button_new();
    gtk_container_add(GTK_CONTAINER(button), icon);
    return button;
}

static void on_webSearch_add(GtkWidget* widget, CBrowser* browser)
{
    GtkWidget* dialog;
    dialog = gtk_dialog_new_with_buttons("Add search engine"
                                , GTK_WINDOW(browser->window)//...
                                , GTK_DIALOG_DESTROY_WITH_PARENT
                                | GTK_DIALOG_NO_SEPARATOR
                                , GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL
                                , GTK_STOCK_ADD, GTK_RESPONSE_ACCEPT
                                , NULL);
    gtk_window_set_icon_name(GTK_WINDOW(dialog), GTK_STOCK_ADD);
    gtk_container_set_border_width(GTK_CONTAINER(dialog), 5);
    gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), 5);
    GtkSizeGroup* sizegroup =  gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
    GtkWidget* hbox = gtk_hbox_new(FALSE, 8);
    gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);
    GtkWidget* label = gtk_label_new_with_mnemonic("_Name:");
    gtk_size_group_add_widget(sizegroup, label);
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    GtkWidget* entry_name = gtk_entry_new();
    gtk_box_pack_start(GTK_BOX(hbox), entry_name, TRUE, TRUE, 0);
    gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), hbox);
    gtk_widget_show_all(hbox);
    
    hbox = gtk_hbox_new(FALSE, 8);
    gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);
    label = gtk_label_new_with_mnemonic("_Url:");
    gtk_size_group_add_widget(sizegroup, label);
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    GtkWidget* entry_url = gtk_entry_new();
    gtk_box_pack_start(GTK_BOX(hbox), entry_url, TRUE, TRUE, 0);
    gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), hbox);
    gtk_widget_show_all(hbox);
    if(gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT)
    {
     CSearchEngine* searchEngine = g_new0(CSearchEngine, 1);
     searchEngine->sShortName = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry_name)));
     searchEngine->sDescription = g_strdup("");//gtk_entry_get_text(GTK_ENTRY(entry_desc));
     searchEngine->sUrl = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry_url)));
     searchEngine->sInputEncoding = g_strdup("UTF-8");//gtk_entry_get_text(GTK_ENTRY(entry_url));
     searchEngine->sIcon = g_strdup(GTK_STOCK_FIND);//gtk_entry_get_text(GTK_ENTRY(entry_icon));
     searchEngine->sKeyword = g_strdup("");//gtk_entry_get_text(GTK_ENTRY(entry_url));
     g_ptr_array_add(searchEngines, searchEngine);
    }
    gtk_widget_destroy(dialog);
}

void on_action_manageSearchEngines_activate(GtkWidget* widget, CBrowser* browser)
{
    // Show the Manage search engines dialog. Create it if necessary.
    static GtkWidget* dialog;
    if(GTK_IS_DIALOG(dialog))
     gtk_window_present(GTK_WINDOW(dialog));
    else
    {
     const gchar* sDialogTitle = "Manage search engines";
     dialog = gtk_dialog_new_with_buttons(sDialogTitle
                                 , GTK_WINDOW(browser->window)
                                 , GTK_DIALOG_DESTROY_WITH_PARENT
                                 | GTK_DIALOG_NO_SEPARATOR
                                 , GTK_STOCK_HELP
                                 , GTK_RESPONSE_HELP
                                 , GTK_STOCK_CLOSE
                                 , GTK_RESPONSE_CLOSE
                                 , NULL);
     gtk_window_set_icon_name(GTK_WINDOW(dialog), GTK_STOCK_PROPERTIES);
     // TODO: Implement some kind of help function
     gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog)
      , GTK_RESPONSE_HELP, FALSE); //...
     guint iWidth, iHeight;
     sokoke_widget_get_text_size(dialog, "M", &iWidth, &iHeight);
     gtk_window_set_default_size(GTK_WINDOW(dialog), iWidth * 45, -1);
     g_signal_connect(dialog, "response"
      , G_CALLBACK(gtk_widget_destroy), dialog);
     // TODO: Do we want tooltips for explainations or can we omit that?
     // TODO: We need mnemonics
     // TODO: Take multiple windows into account when applying changes
     GtkWidget* xfce_heading;
     if(xfce_heading = sokoke_xfce_header_new(
      gtk_window_get_icon_name(GTK_WINDOW(dialog)), sDialogTitle))
      gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox)
      , xfce_heading, FALSE, FALSE, 0);
     GtkWidget* hbox = gtk_hbox_new(FALSE, 0);
     gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), hbox, TRUE, TRUE, 12);
     GtkTreeViewColumn* column;
     GtkCellRenderer* renderer_text; GtkCellRenderer* renderer_pixbuf;
     GtkListStore* liststore = gtk_list_store_new(ENGINES_COL_N
      , G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
     GtkWidget* treeview = gtk_tree_view_new_with_model(GTK_TREE_MODEL(liststore));
     gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(treeview), FALSE);
     column = gtk_tree_view_column_new();
     renderer_pixbuf = gtk_cell_renderer_pixbuf_new();
     gtk_tree_view_column_pack_start(column, renderer_pixbuf, FALSE);
     gtk_tree_view_column_set_cell_data_func(column, renderer_pixbuf
     , (GtkTreeCellDataFunc)on_webSearch_engines_render_icon, treeview, NULL);
     renderer_text = gtk_cell_renderer_text_new();
     gtk_tree_view_column_pack_start(column, renderer_text, TRUE);
     gtk_tree_view_column_set_cell_data_func(column, renderer_text
     , (GtkTreeCellDataFunc)on_webSearch_engines_render_text, treeview, NULL);
     gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);
     GtkWidget* scrolled = gtk_scrolled_window_new(NULL, NULL);
     gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled)
     , GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
     gtk_container_add(GTK_CONTAINER(scrolled), treeview);
     gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolled), GTK_SHADOW_IN);
     gtk_box_pack_start(GTK_BOX(hbox), scrolled, TRUE, TRUE, 5);
     guint i; for(i = 0; i < searchEngines->len; i++)
     {
      CSearchEngine* searchEngine = (CSearchEngine*)g_ptr_array_index(searchEngines, i);
      gtk_list_store_insert_with_values(GTK_LIST_STORE(liststore), NULL, i
       , ENGINES_COL_ICON, searchEngine->sIcon
       , ENGINES_COL_NAME, searchEngine->sShortName
       , ENGINES_COL_DESCRIPTION , searchEngine->sDescription
       , -1);
     }
     g_object_unref(liststore);
     GtkWidget* vbox = gtk_vbox_new(FALSE, 4);
     gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 4);
     GtkWidget* button = mini_button_new_from_stock(GTK_STOCK_ADD);
     g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(on_webSearch_add), browser);
     gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
     button = mini_button_new_from_stock(GTK_STOCK_EDIT);
     gtk_widget_set_sensitive(button, FALSE); //...
     gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
     gtk_widget_set_sensitive(button, FALSE); //...
     button = mini_button_new_from_stock(GTK_STOCK_REMOVE);
     gtk_widget_set_sensitive(button, FALSE); //...
     gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
     button = gtk_label_new(""); // This is an invisible separator
     gtk_box_pack_start(GTK_BOX(vbox), button, TRUE, TRUE, 12);
     button = mini_button_new_from_stock(GTK_STOCK_GO_DOWN);
     gtk_widget_set_sensitive(button, FALSE); //...
     gtk_box_pack_end(GTK_BOX(vbox), button, FALSE, FALSE, 0);
     button = mini_button_new_from_stock(GTK_STOCK_GO_UP);
     gtk_widget_set_sensitive(button, FALSE); //...
     gtk_box_pack_end(GTK_BOX(vbox), button, FALSE, FALSE, 0);
     gtk_widget_show_all(GTK_DIALOG(dialog)->vbox);
     gtk_widget_show(dialog);
    }
}

gboolean on_webSearch_key_down(GtkWidget* widget
 , GdkEventKey* event, CBrowser* browser)
{
    switch(event->keyval)
    {
     case GDK_Up:
      update_searchEngine(config->iSearchEngine - 1, browser);
      return TRUE;
     case GDK_Down:
      update_searchEngine(config->iSearchEngine + 1, browser);
      return TRUE;
    }
    return FALSE;
}

gboolean on_webSearch_scroll(GtkWidget* document, GdkEventScroll* event
 , CBrowser* browser)
{
    if(event->direction == GDK_SCROLL_DOWN)
     update_searchEngine(config->iSearchEngine + 1, browser);
    else if(event->direction == GDK_SCROLL_UP)
     update_searchEngine(config->iSearchEngine - 1, browser);
    return TRUE;
}

void on_webSearch_activate(GtkWidget* widget, CBrowser* browser)
{
    const gchar* sKeywords = gtk_entry_get_text(GTK_ENTRY(widget));
    gchar* sUrl;
    if(searchEngines->len)
    {
     CSearchEngine* searchEngine = (CSearchEngine*)g_ptr_array_index(searchEngines, config->iSearchEngine);
     sUrl = searchEngine->sUrl;
    }
    else // The location search is our fallback
     sUrl = config->sLocationSearch;
    gchar* sSearch;
    if(strstr(sUrl, "%s"))
     sSearch = g_strdup_printf(sUrl, sKeywords);
    else
     sSearch = g_strconcat(sUrl, " ", sKeywords, NULL);
    entry_completion_append(GTK_ENTRY(widget), sKeywords);
    document_load_uri(get_current_document(browser), sSearch);
    g_free(sSearch);
}
